import json
import logging
import re
from json import loads, dumps
from typing import Optional, Union, Dict, List, Any, overload, Literal, Tuple
from urllib.parse import urlparse, urljoin
from warnings import warn

from httpx import InvalidURL, URL, HTTPStatusError

from ipfabric.api import IPFabricAPI
from ipfabric.diagrams import Diagram
from ipfabric.models import Technology, Inventory, Jobs, Intent, Devices
from ipfabric.tools import TIMEZONES

try:
    from pandas import DataFrame
except ImportError:
    DataFrame = None

logger = logging.getLogger("ipfabric")

RE_PATH = re.compile(r"^/?(api/)?v\d(\.\d)?/")
RE_TABLE = re.compile(r"^tables/")
EXPORT_FORMAT = Literal["json", "csv", "df"]


class IPFClient(IPFabricAPI):
    def __init__(
        self,
        base_url: Optional[str] = None,
        api_version: Optional[str] = None,
        snapshot_id: Optional[str] = None,
        auth: Optional[Any] = None,
        unloaded: bool = False,
        env_file: Optional[str] = None,
        streaming: bool = True,
        **kwargs,
    ):
        """Initializes the IP Fabric Client

        Args:
            base_url: IP Fabric instance provided in 'base_url' parameter, or the 'IPF_URL' environment variable
            api_version: [Optional] Version of IP Fabric API
            auth: API token, tuple (username, password), or custom Auth to pass to httpx
            snapshot_id: IP Fabric snapshot ID to use by default for database actions - defaults to '$last'
            unloaded: True to load metadata from unloaded snapshots
            env_file: Path to .env file to load
            streaming: Stream table requests instead of paging results, see documentation for important information.
            **kwargs: Keyword args to pass to httpx
        """
        super().__init__(
            base_url=base_url,
            api_version=api_version,
            auth=auth,
            snapshot_id=snapshot_id,
            unloaded=unloaded,
            env_file=env_file,
            streaming=streaming,
            **kwargs,
        )
        self.inventory = Inventory(client=self)
        self.technology = Technology(client=self)
        self.jobs = Jobs(client=self)
        self.diagram = Diagram(ipf=self)
        self.intent = Intent(client=self, load=False)
        self._devices = list()

    @property
    def snapshot_id(self) -> str:
        """get snapshot Id"""
        return self._snapshot_id

    @snapshot_id.setter
    def snapshot_id(self, snapshot_id):
        super(self.__class__, self.__class__).snapshot_id.fset(self, snapshot_id)
        if not getattr(self, "inventory", None):
            self.inventory = Inventory(client=self)
        if self.snapshot.disabled_intent_verification is False and getattr(self, "intent", None) and self.intent.loaded:
            self.intent = Intent(client=self)
        if getattr(self, "_devices", None):
            self._devices = self.load_devices()

    def update(self):
        super(self.__class__, self.__class__).update(self)
        if self.snapshot_id not in self.snapshots:
            logger.warning(f"Snapshot {self.snapshot_id} is no longer loaded switching to `$last`.")
        self.snapshot_id = self.snapshot_id if self.snapshot_id in self.snapshots else "$last"

    @property
    def devices(self) -> Devices:
        """get devices"""
        if not self._devices:
            logger.info("Devices not loaded, loading devices.")
            self._devices = self.load_devices()
        return self._devices

    @devices.setter
    def devices(self, devices):
        self._devices = devices

    def load_devices(self, device_filters: dict = None, device_attr_filters: dict = None):
        self.devices = Devices(
            client=self,
            snapshot_id=self.snapshot_id,
            device_filters=device_filters,
            device_attr_filters=device_attr_filters,
        )
        return self.devices

    def _check_url(self, url):
        path = urlparse(url).path
        path = path if path[0] == "/" else "/" + path
        if path in self.web_to_api:
            return self.web_to_api[path].api_endpoint
        r = RE_PATH.search(path)
        url = path[r.end():] if r else path  # fmt: skip
        url = url[1:] if url[0] == "/" else url
        return url

    def _create_payload(self, url, snapshot_id, filters, sort, attr_filters):
        payload = dict()
        if self.api_version >= "v6.3":  # TODO: No check in v7.0
            payload = {"format": {"dataType": "json"}}
        if filters and isinstance(filters, str):
            payload["filters"] = loads(filters)
        elif filters and isinstance(filters, dict):
            payload["filters"] = filters
        if snapshot_id:
            payload["snapshot"] = snapshot_id
        if sort:
            payload["sort"] = sort
        if RE_TABLE.match(url) and (attr_filters or self.attribute_filters):
            payload["attributeFilters"] = attr_filters or self.attribute_filters
        return payload

    def _check_url_payload(self, url, snapshot_id, filters, reports, sort, attr_filters, export, csv_tz):
        url = self._check_url(url)
        payload = self._create_payload(url, snapshot_id, filters, sort, attr_filters)

        if export != "csv":
            if isinstance(reports, (str, list)):
                payload["reports"] = reports
            elif reports is True and self.oas.get(url, None) and self.oas[url].web_endpoint:
                payload["reports"] = self.oas[url].web_endpoint
            elif reports is True and (not self.oas.get(url, None) or not self.oas[url].web_endpoint):
                logger.warning(
                    f"Could not automatically discover Web Endpoint for Intent Data for table '/{url}'.\n"
                    f"Returning results without Intent Rules."
                )
        else:
            if reports:
                logger.warning("CSV export does not return reports, parameter has been excluded.")
            payload["format"] = {"exportToFile": True, "dataType": "csv"}
            if csv_tz and csv_tz.lower() not in TIMEZONES:
                raise ValueError(
                    f"CSV timezone '{csv_tz}' not in available timezones; see `ipfabric.tools.shared.TIMEZONES`"
                )
            elif csv_tz:
                payload["format"]["options"] = {"timezone": TIMEZONES[csv_tz.lower()]}
        return url, payload

    def _get_payload(self, url, payload):
        snapshot_id = payload.pop("snapshot", None)
        reports = payload.pop("reports") if "reports" in payload and isinstance(payload["reports"], str) else None
        p = "&".join([f"{k}={dumps(v, separators=(',', ':'))}" for k, v in payload.items()])
        if snapshot_id:
            p += f"&snapshot={snapshot_id}"
        if reports:
            p += f"&reports={reports}"
        url = urljoin(str(self.base_url), url + f"?{p}")
        if len(url) > 4096:
            return False
        return url

    def _stream(self, url, payload, export):
        if export == "csv" and self.api_version < "v6.3":
            raise NotImplementedError(f"CSV export is implemented in v6.3.0; IPF version: {self.os_version}.")
        get_url = self._get_payload(url, payload)
        if get_url is False and export == "csv":
            raise InvalidURL("URL exceeds max character limit of 4096 cannot export to CSV.")
        elif get_url is False:
            logger.warning("URL exceeds max character limit of 4096 switching to pagination.")
            return False
        with self.stream("GET", get_url) as stream_resp:
            stream_resp.raise_for_status()
            data = stream_resp.read()
        return data if export == "csv" else loads(data)["data"]

    def query(self, url: str, payload: Union[str, dict], get_all: bool = True) -> Union[List[dict], bytes]:
        """Submits a query, does no formatting on the parameters.  Use for copy/pasting from the webpage.

        Args:
            url: Example: https://demo1.ipfabric.io/api/v1/tables/vlan/device-summary or tables/vlan/device-summary
            payload: Dictionary to submit in POST or can be JSON string (i.e. read from file).
            get_all: Default use pager to get all results and ignore pagination information in the payload

        Returns:
            list or bytes: List of Dictionary objects or bytes if csv.
        """
        url = self._check_url(url)
        if isinstance(payload, str):
            payload = loads(payload)
        export = payload.get("format", {}).get("dataType", "json")
        data = False
        if export == "csv" or (self.streaming and get_all):
            data = self._stream(url, payload, export)
        elif get_all and data is False:
            data = self._ipf_pager(url, payload)
        elif data is False:
            res = self.post(url, json=payload)
            res.raise_for_status()
            data = res.json()["data"]
        return data

    def _get_columns(self, url: str):  # TODO: Remove in v6.8
        warn("This method will be removed in v7.0, please use `get_columns`.", DeprecationWarning, stacklevel=2)
        logger.warning("This method will be removed in v7.0, please use `get_columns`.")
        return self.get_columns(url=url)

    def get_columns(self, url: str) -> List[str]:
        """Checks OAS to find available columns.

        Args:
            url: API url to post

        Returns:
            list: List of column names
        """
        if self.oas.get(url, None) and self.oas[url].columns:
            return self.oas[url].columns
        else:
            raise LookupError(f"Could not find columns for table '/{url}'.")

    def get_count(
        self,
        url: str,
        filters: Optional[Union[dict, str]] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        snapshot: bool = True,
    ) -> int:
        """Get a total number of rows
        Args:
            url: API URL to post to
            filters: Optional dictionary of filters
            attr_filters: Optional dictionary of attribute filters
            snapshot_id: Optional snapshot_id to override default
            snapshot: Set to False for some tables like management endpoints.
        Returns:
            int: a count of rows
        """
        snapshot_id = snapshot_id or self.snapshot_id if snapshot else None
        url, payload = self._check_url_payload(url, snapshot_id, filters, None, None, attr_filters, None, None)
        payload.update({"columns": ["id"], "pagination": {"limit": 1, "start": 0}})
        res = self.post(url, json=payload)
        res.raise_for_status()
        return res.json()["_meta"]["count"]

    def _fetch_setup(self, url, export, columns, snapshot_id, filters, reports, sort, attr_filters, csv_tz, snapshot):
        if export == "df" and DataFrame is None:
            raise ImportError("pandas not installed. Run `pip install ipfabric[pd]`.")
        snapshot_id = snapshot_id or self.snapshot_id if snapshot else None
        if "color" in json.dumps(filters) and not reports:
            reports = True
        url, payload = self._check_url_payload(url, snapshot_id, filters, reports, sort, attr_filters, export, csv_tz)
        payload["columns"] = columns or self.get_columns(url)
        return url, payload

    @overload
    def fetch(
        self,
        url,
        export: Literal["json"] = ...,
        columns: Optional[List] = None,
        filters: Optional[Union[dict, str]] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot: bool = True,
    ) -> List[dict]: ...

    @overload
    def fetch(
        self,
        url,
        export: Literal["csv"],
        columns: Optional[List] = None,
        filters: Optional[Union[dict, str]] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot: bool = True,
        csv_tz: Optional[str] = None,
    ) -> bytes: ...

    @overload
    def fetch(
        self,
        url,
        export: Literal["df"] = ...,
        columns: Optional[List] = None,
        filters: Optional[Union[dict, str]] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot: bool = True,
    ) -> DataFrame: ...

    def fetch(
        self,
        url,
        export: EXPORT_FORMAT = "json",
        columns: Optional[List] = None,
        filters: Optional[Union[dict, str]] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot: bool = True,
        csv_tz: Optional[str] = None,
    ):
        """Gets data from IP Fabric for specified endpoint

        Args:
            url: Example tables/vlan/device-summary
            export: str: Export format to return [json, csv]; default is json.
            columns: Optional list of columns to return, None will return all
            filters: Optional dictionary of filters
            limit: Default to 1,000 rows
            start: Starts at 0
            snapshot_id: Optional snapshot_id to override default
            reports: String of frontend URL where the reports are displayed or a list of report IDs
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            attr_filters: Optional dictionary to apply an Attribute filter
            snapshot: Set to False for some tables like management endpoints.
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
        Returns:
            Union[List[dict], bytes, pandas.DataFrame]: List of dict if json, bytes string if CSV, DataFrame is df
        """
        url, payload = self._fetch_setup(
            url, export, columns, snapshot_id, filters, reports, sort, attr_filters, csv_tz, snapshot
        )
        payload["pagination"] = dict(start=start, limit=limit)
        data = False
        if export == "csv" or self.streaming:
            data = self._stream(url, payload, export)
        if data is False:
            res = self.post(url, json=payload)
            res.raise_for_status()
            data = res.json()["data"]
        if export == "df":
            data = DataFrame.from_records(data) if export == "df" else data
        return data

    @overload
    def fetch_all(
        self,
        url: str,
        export: Literal["json"] = ...,
        columns: Optional[List] = None,
        filters: Optional[Union[dict, str]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot: bool = True,
    ) -> List[dict]: ...

    @overload
    def fetch_all(
        self,
        url: str,
        export: Literal["csv"],
        columns: Optional[List] = None,
        filters: Optional[Union[dict, str]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot: bool = True,
        csv_tz: Optional[str] = None,
    ) -> bytes: ...

    @overload
    def fetch_all(
        self,
        url: str,
        export: Literal["df"],
        columns: Optional[List] = None,
        filters: Optional[Union[dict, str]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot: bool = True,
    ) -> DataFrame: ...

    def fetch_all(
        self,
        url: str,
        export: EXPORT_FORMAT = "json",
        columns: Optional[List] = None,
        filters: Optional[Union[dict, str]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot: bool = True,
        csv_tz: Optional[str] = None,
    ):
        """Gets all data from IP Fabric for specified endpoint

        Args:
            url: Example tables/vlan/device-summary
            export: str: Export format to return [json, csv]; default is json.
            columns: Optional list of columns to return, None will return all
            filters: Optional dictionary of filters
            snapshot_id: Optional snapshot_id to override default
            reports: String of frontend URL where the reports are displayed or a list of report IDs
            sort: Optional dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            attr_filters: Optional dictionary to apply an Attribute filter
            snapshot: Set to False for some tables like management endpoints.
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
        Returns:
            Union[List[dict], bytes, pandas.DataFrame]: List of dict if json, bytes string if CSV, DataFrame is df
        """
        url, payload = self._fetch_setup(
            url, export, columns, snapshot_id, filters, reports, sort, attr_filters, csv_tz, snapshot
        )
        data = False
        if export == "csv" or self.streaming:
            data = self._stream(url, payload, export)
        if data is False:
            data = self._ipf_pager(url, payload)
        if export == "df":
            data = DataFrame.from_records(data)
        return data

    @staticmethod
    def _shared_url_id(url: Union[int, str]) -> Tuple[str, str]:
        try:
            return str(int(url)), None
        except ValueError:
            parsed_url = URL(url)
            if parsed_url.params.get("copyId"):
                return parsed_url.params.get("copyId"), "table"
            else:
                return parsed_url.path.split("/")[-1], "graph"

    def _get_shared_view(self, url_id: str, view: str = None) -> dict:
        if view == "table":
            resp = self.get(f"/tables/url/{url_id}")
        elif view == "graph":
            resp = self.get(f"/graphs/urls/{url_id}")
        else:
            try:
                resp = self.get(f"/tables/url/{url_id}")
                resp.raise_for_status()
            except HTTPStatusError:
                resp = self.get(f"/graphs/urls/{url_id}")
                resp.raise_for_status()
        return resp.json()

    def _table_url(self, url_id: str, data=True, reports=True):
        query = self._get_shared_view(url_id, "table")
        columns = list(set(query.get("columnVisibility", {}).keys()).union(set(query.get("columnWidth", {}).keys())))
        filters, sort = query.get("filters", {}), query.get("sort", {})
        filters.pop("selectedFilter", None)
        url = self.web_to_api[query["webEndpoint"]].api_endpoint

        code = f"{url.split('/')[-1]} = ipf.fetch_all(\n    {url=},\n    {columns=},"
        code += f"\n    {filters=}," if filters else ""
        code += f"\n    {sort=}," if sort else ""
        code += f"\n    reports='{query['webEndpoint']}',\n)" if reports else "\n)"

        if data:
            logger.debug(code)
            return self.fetch_all(
                url=url, columns=columns, filters=filters, sort=sort, reports=query["webEndpoint"] if reports else False
            )
        return code

    def shared_view(self, url: Union[int, str], data=True, reports: bool = True) -> Union[str, list]:
        """Takes a shared table view link and returns the data or the code to implement in python.

        Args:
            url: Id of the shared view (1453653298) or full/partial URL (`/inventory/devices?copyId=1453653298`)
            data: Defaults to return the data instead of printing the code
            reports: True to return Intent Verification data

        Returns: List of dictionaries of the data or a string representing the code.
        """
        url_id, _ = self._shared_url_id(url)
        return self._table_url(url_id, data, reports)

    def shared_url(self, url: Union[int, str], data=True) -> Union[str, list, dict]:
        """Takes a shared link and returns the data or the code to implement in python.

        Args:
            url: Id of the shared view (1453653298) or
                 full/partial URL (`/inventory/devices?copyId=1453653298` or `/diagrams/share/1453626097`)
            data: Defaults to return the data instead of printing the code

        Returns: Data (list of dicts for table or dict for graph) or a string representing the code.
        """
        url_id, view = self._shared_url_id(url)
        if view == "table":
            return self.shared_view(url_id, data)
        elif view == "graph":
            return self.diagram.shared_view(url_id, data="json" if data else "code")
        if "webEndpoint" in self._get_shared_view(url_id):
            return self.shared_view(url_id, data)
        return self.diagram.shared_view(url_id, data="json" if data else "code")



import logging
from typing import Union, Dict, List, Optional, Literal
from warnings import warn

from ipfabric.api import IPFabricAPI
from .input_models.graph_parameters import Unicast, Multicast, Host2GW, Network
from .input_models.graph_settings import (
    NetworkSettings,
    PathLookupSettings,
    GraphSettings,
    Overlay,
    GroupSettings,
)
from .output_models.graph_result import NetworkEdge, Node, PathLookupEdge, GraphResult, PathLookup

logger = logging.getLogger("ipfabric")
WARNING = "This method will be removed in v6.9."
GRAPHS_URL = "graphs/"


class Diagram:
    def __init__(self, ipf):
        self.ipf = ipf

    def _check_graph_cache(self, snapshot_id):
        snapshot_id = snapshot_id or self.ipf.snapshot_id
        if snapshot_id not in self.ipf.loaded_snapshots:
            raise ValueError(f"Snapshot {snapshot_id} is not loaded or not found in IP Fabric.")
        if self.ipf.snapshots[snapshot_id].disabled_graph_cache:
            raise ValueError(f"Snapshot {snapshot_id} has Graph Cache tasks disabled.")
        return snapshot_id

    def _create_overlay(self, overlay: dict) -> Overlay:
        if overlay.get("intentRuleId", None):
            try:
                Overlay._valid_intentrule(overlay["intentRuleId"])
            except ValueError:
                if not self.ipf.intent.loaded:
                    self.ipf.intent.load_intent()
                intents = self.ipf.intent.intents_by_name[overlay["intentRuleId"]]
                if len(intents) > 1:
                    raise ValueError(f"Multiple Intents found with name `{overlay['intentRuleId']}`.")
                else:
                    overlay["intentRuleId"] = intents[0].intent_id
        return Overlay(**overlay)

    def _format_overlay(self, overlay: Union[Overlay, dict], snapshot_id: str = None) -> dict:
        if isinstance(overlay, dict):
            overlay = self._create_overlay(overlay)

        if overlay.type == "compare":
            self._check_graph_cache(overlay.snapshotToCompare)
            overlay.snapshotToCompare = self.ipf.snapshots[overlay.snapshotToCompare].snapshot_id
            if snapshot_id or self.ipf.snapshot_id == overlay.snapshotToCompare:
                raise ValueError(f"Cannot compare snapshot `{overlay.snapshotToCompare}` to itself.")
        return overlay.overlay()

    def _query(
        self,
        parameters: dict,
        snapshot_id: str = None,
        overlay: Union[Overlay, dict] = None,
        image: str = None,
        graph_settings: Union[NetworkSettings, PathLookupSettings, GraphSettings] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
    ):
        """
        Submits a query, does no formatting on the parameters.  Use for copy/pasting from the webpage.
        :param parameters: dict: Dictionary to submit in POST.
        :return: list: List of Dictionary objects.
        """
        url = GRAPHS_URL + image if image else GRAPHS_URL
        payload = dict(parameters=parameters, snapshot=self._check_graph_cache(snapshot_id))
        if overlay:
            payload["overlay"] = self._format_overlay(overlay, snapshot_id)
        if graph_settings:
            payload["settings"] = graph_settings.settings()
        if attr_filters or self.ipf.attribute_filters:
            payload["attributeFilters"] = attr_filters or self.ipf.attribute_filters
        res = self.ipf.post(url, json=payload)
        res.raise_for_status()
        return res.content if image else res.json()

    def json(
        self,
        parameters: Union[Unicast, Multicast, Host2GW, Network],
        snapshot_id: str = None,
        overlay: Union[Overlay, dict] = None,
        graph_settings: Union[NetworkSettings, PathLookupSettings, GraphSettings] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        unicast_swap_src_dst: bool = False,
    ) -> dict:
        return self._query(
            parameters.parameters(unicast_swap_src_dst) if isinstance(parameters, Unicast) else parameters.parameters(),
            snapshot_id=snapshot_id,
            overlay=overlay,
            attr_filters=attr_filters,
            graph_settings=graph_settings,
        )

    def svg(
        self,
        parameters: Union[Unicast, Multicast, Host2GW, Network],
        snapshot_id: str = None,
        overlay: Union[Overlay, dict] = None,
        graph_settings: Union[NetworkSettings, PathLookupSettings, GraphSettings] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        unicast_swap_src_dst: bool = False,
    ) -> bytes:
        return self._query(
            parameters.parameters(unicast_swap_src_dst) if isinstance(parameters, Unicast) else parameters.parameters(),
            snapshot_id=snapshot_id,
            overlay=overlay,
            attr_filters=attr_filters,
            image="svg",
            graph_settings=graph_settings,
        )

    def png(
        self,
        parameters: Union[Unicast, Multicast, Host2GW, Network],
        snapshot_id: str = None,
        overlay: Union[Overlay, dict] = None,
        graph_settings: Union[NetworkSettings, PathLookupSettings, GraphSettings] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        unicast_swap_src_dst: bool = False,
    ) -> bytes:
        return self._query(
            parameters.parameters(unicast_swap_src_dst) if isinstance(parameters, Unicast) else parameters.parameters(),
            snapshot_id=snapshot_id,
            overlay=overlay,
            attr_filters=attr_filters,
            image="png",
            graph_settings=graph_settings,
        )

    def share_link(
        self,
        parameters: Union[Unicast, Multicast, Host2GW, Network],
        snapshot_id: str = None,
        overlay: Union[Overlay, dict] = None,
        graph_settings: Union[NetworkSettings, PathLookupSettings, GraphSettings] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        unicast_swap_src_dst: bool = False,
    ) -> str:
        parameters = (
            parameters.parameters(unicast_swap_src_dst) if isinstance(parameters, Unicast) else parameters.parameters()
        )
        resp = self._query(
            parameters,
            snapshot_id=snapshot_id,
            overlay=overlay,
            attr_filters=attr_filters,
            graph_settings=graph_settings,
        )

        parameters.pop("layouts", None)
        payload = {
            "graphView": {
                "name": "Shared view",
                "parameters": parameters,
                "collapsedNodeGroups": [],
                "hiddenNodes": [],
                "positions": {k: v["position"] for k, v in resp["graphResult"]["graphData"]["nodes"].items()},
                "settings": resp["graphResult"]["settings"],
            },
            "snapshot": self._check_graph_cache(snapshot_id),
        }
        if overlay:
            payload["graphView"]["overlay"] = self._format_overlay(overlay, snapshot_id)
        res = self.ipf.post("graphs/urls", json=payload)
        res.raise_for_status()
        return str(self.ipf.base_url.join(f"/diagrams/share/{res.json()['id']}"))

    def model(
        self,
        parameters: Union[Unicast, Multicast, Host2GW, Network],
        snapshot_id: str = None,
        overlay: Overlay = None,
        graph_settings: Union[NetworkSettings, PathLookupSettings, GraphSettings] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        unicast_swap_src_dst: bool = False,
    ) -> GraphResult:
        json_data = self.json(parameters, snapshot_id, overlay, graph_settings, attr_filters, unicast_swap_src_dst)
        edge_setting_dict = self._diagram_edge_settings(json_data["graphResult"]["settings"])
        if isinstance(parameters, Network):
            return self._diagram_network(json_data, edge_setting_dict)
        else:
            return self._diagram_pathlookup(json_data, edge_setting_dict)

    @staticmethod
    def _diagram_network(json_data: dict, edge_setting_dict: dict, pathlookup: bool = False) -> GraphResult:
        edges, nodes = dict(), dict()
        for node_id, node in json_data["graphResult"]["graphData"]["nodes"].items():
            nodes[node_id] = Node.model_validate(node)

        for edge_id, edge_json in json_data["graphResult"]["graphData"]["edges"].items():
            edge = PathLookupEdge.model_validate(edge_json) if pathlookup else NetworkEdge.model_validate(edge_json)
            edge.protocol = (
                edge_setting_dict[edge.edgeSettingsId].name if edge.edgeSettingsId in edge_setting_dict else None
            )
            if edge.source:
                edge.source = nodes[edge.source]
            if edge.target:
                edge.target = nodes[edge.target]
            edges[edge_id] = edge

        return GraphResult(edges=edges, nodes=nodes)

    def _diagram_pathlookup(self, json_data: dict, edge_setting_dict: dict) -> GraphResult:
        graph_result = self._diagram_network(json_data, edge_setting_dict, pathlookup=True)

        for edge_id, edge in graph_result.edges.items():
            for prev_id in edge.prevEdgeIds:
                edge.prevEdge.append(graph_result.edges[prev_id])
            for next_id in edge.nextEdgeIds:
                edge.nextEdge.append(graph_result.edges[next_id] if next_id in graph_result.edges else next_id)
        graph_result.pathlookup = PathLookup.model_validate(json_data["pathlookup"])

        return graph_result

    @staticmethod
    def _diagram_edge_settings(graph_settings: dict) -> dict:
        net_settings = GraphSettings(**graph_settings)
        edge_setting_dict = dict()
        for edge in net_settings.edges:
            edge_setting_dict[edge.id] = edge
            if isinstance(edge, GroupSettings):
                for child in edge.children:
                    edge_setting_dict[child.id] = child
        return edge_setting_dict

    # TODO: Remove in v7.0
    def diagram_json(
        self,
        parameters: Union[Unicast, Multicast, Host2GW, Network],
        snapshot_id: str = None,
        overlay: Overlay = None,
        graph_settings: Union[NetworkSettings, PathLookupSettings, GraphSettings] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        unicast_swap_src_dst: bool = False,
    ) -> dict:
        warn(f"{WARNING} Please use IPFClient().diagram.json()", DeprecationWarning, stacklevel=2)
        logger.warning(f"{WARNING} Please use IPFClient().diagram.json()")
        return self._query(
            parameters.parameters(unicast_swap_src_dst) if isinstance(parameters, Unicast) else parameters.parameters(),
            snapshot_id=snapshot_id,
            overlay=overlay,
            attr_filters=attr_filters,
            graph_settings=graph_settings,
        )

    def diagram_svg(
        self,
        parameters: Union[Unicast, Multicast, Host2GW, Network],
        snapshot_id: str = None,
        overlay: Overlay = None,
        graph_settings: Union[NetworkSettings, PathLookupSettings, GraphSettings] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        unicast_swap_src_dst: bool = False,
    ) -> bytes:
        warn(f"{WARNING} Please use IPFClient().diagram.svg()", DeprecationWarning, stacklevel=2)
        logger.warning(f"{WARNING} Please use IPFClient().diagram.json()")
        return self._query(
            parameters.parameters(unicast_swap_src_dst) if isinstance(parameters, Unicast) else parameters.parameters(),
            snapshot_id=snapshot_id,
            overlay=overlay,
            attr_filters=attr_filters,
            image="svg",
            graph_settings=graph_settings,
        )

    def diagram_png(
        self,
        parameters: Union[Unicast, Multicast, Host2GW, Network],
        snapshot_id: str = None,
        overlay: Overlay = None,
        graph_settings: Union[NetworkSettings, PathLookupSettings, GraphSettings] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        unicast_swap_src_dst: bool = False,
    ) -> bytes:
        warn(f"{WARNING} Please use IPFClient().diagram.png()", DeprecationWarning, stacklevel=2)
        logger.warning(f"{WARNING} Please use IPFClient().diagram.json()")
        return self._query(
            parameters.parameters(unicast_swap_src_dst) if isinstance(parameters, Unicast) else parameters.parameters(),
            snapshot_id=snapshot_id,
            overlay=overlay,
            attr_filters=attr_filters,
            image="png",
            graph_settings=graph_settings,
        )

    def diagram_model(
        self,
        parameters: Union[Unicast, Multicast, Host2GW, Network],
        snapshot_id: str = None,
        overlay: Overlay = None,
        graph_settings: Union[NetworkSettings, PathLookupSettings, GraphSettings] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        unicast_swap_src_dst: bool = False,
    ) -> GraphResult:
        warn(f"{WARNING} Please use IPFClient().diagram.json()", DeprecationWarning, stacklevel=2)
        logger.warning(f"{WARNING} Please use IPFClient().diagram.json()")
        return self.model(
            parameters,
            snapshot_id=snapshot_id,
            overlay=overlay,
            graph_settings=graph_settings,
            attr_filters=attr_filters,
            unicast_swap_src_dst=unicast_swap_src_dst,
        )

    def _graph_url(self, url_id: str, data: Literal["json", "code", "model", "svg", "png"] = "json"):
        query = self.ipf._get_shared_view(url_id, "graph")

        code = ""
        if data == "code":
            return code
        logger.debug(code)
        raise NotImplementedError("WIP.")

    def shared_view(self, url: Union[int, str], data: Literal["json", "code", "model", "svg", "png"] = "json"):
        """Takes a shared graph link and returns the data or the code to implement in python.

        Args:
            url: Id of the shared view (1453653298) or full/partial URL (`/diagrams/share/1453626097`)
            data: Defaults to return the data instead of printing the code

        Returns: The graph data or string representing the code to produce it.
        """
        url_id, _ = self.ipf._shared_url_id(url)
        return self._graph_url(url_id, data)


class IPFDiagram(Diagram):
    def __init__(self, ipf=None, **kwargs):
        warn(f"{self.__class__.__name__} will be removed in v6.9.", DeprecationWarning, stacklevel=2)
        logger.warning(f"{self.__class__.__name__} will be removed in v6.9.")
        if ipf:
            super().__init__(ipf)
        else:
            kwargs["timeout"] = kwargs.get("timeout", 15)
            super().__init__(IPFabricAPI(**kwargs))



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Oam(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def unidirectional_link_detection_neighbors(self):
        return models.Table(
            client=self.client, endpoint="tables/management/oam/unidirectional-link-detection/neighbors", sn=self.sn
        )

    @property
    def unidirectional_link_detection_interfaces(self):
        return models.Table(
            client=self.client, endpoint="tables/management/oam/unidirectional-link-detection/interfaces", sn=self.sn
        )



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Sdwan(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def sites(self):
        return models.Table(client=self.client, endpoint="tables/sdwan/sites", sn=self.sn)

    @property
    def links(self):
        return models.Table(client=self.client, endpoint="tables/sdwan/links", sn=self.sn)



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Vlans(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def device_summary(self):
        return models.Table(client=self.client, endpoint="tables/vlan/device-summary", sn=self.sn)

    @property
    def device_detail(self):
        return models.Table(client=self.client, endpoint="tables/vlan/device", sn=self.sn)

    @property
    def network_summary(self):
        return models.Table(client=self.client, endpoint="tables/vlan/network-summary", sn=self.sn)

    @property
    def site_summary(self):
        return models.Table(client=self.client, endpoint="tables/vlan/site-summary", sn=self.sn)

    @property
    def l3_gateways(self):
        return models.Table(client=self.client, endpoint="tables/vlan/l3-gateways", sn=self.sn)



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Management(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def aaa_servers(self):
        return models.Table(client=self.client, endpoint="tables/security/aaa/servers", sn=self.sn)

    @property
    def aaa_lines(self):
        return models.Table(client=self.client, endpoint="tables/security/aaa/lines", sn=self.sn)

    @property
    def aaa_authentication(self):
        return models.Table(client=self.client, endpoint="tables/security/aaa/authentication", sn=self.sn)

    @property
    def aaa_authorization(self):
        return models.Table(client=self.client, endpoint="tables/security/aaa/authorization", sn=self.sn)

    @property
    def aaa_accounting(self):
        return models.Table(client=self.client, endpoint="tables/security/aaa/accounting", sn=self.sn)

    @property
    def aaa_users(self):
        return models.Table(client=self.client, endpoint="tables/security/aaa/users", sn=self.sn)

    @property
    def aaa_password_strength(self):
        return models.Table(client=self.client, endpoint="tables/security/aaa/password-strength", sn=self.sn)

    @property
    def telnet_access(self):
        return models.Table(client=self.client, endpoint="tables/security/enabled-telnet", sn=self.sn)

    @property
    def saved_config_consistency(self):
        return models.Table(client=self.client, endpoint="tables/management/configuration/saved", sn=self.sn)

    @property
    def ntp_summary(self):
        return models.Table(client=self.client, endpoint="tables/management/ntp/summary", sn=self.sn)

    @property
    def ntp_sources(self):
        return models.Table(client=self.client, endpoint="tables/management/ntp/sources", sn=self.sn)

    @property
    def port_mirroring(self):
        return models.Table(client=self.client, endpoint="tables/management/port-mirroring", sn=self.sn)

    @property
    def logging_summary(self):
        return models.Table(client=self.client, endpoint="tables/management/logging/summary", sn=self.sn)

    @property
    def logging_remote(self):
        return models.Table(client=self.client, endpoint="tables/management/logging/remote", sn=self.sn)

    @property
    def logging_local(self):
        return models.Table(client=self.client, endpoint="tables/management/logging/local", sn=self.sn)

    @property
    def flow_overview(self):
        return models.Table(client=self.client, endpoint="tables/management/flow/overview", sn=self.sn)

    @property
    def netflow_devices(self):
        return models.Table(client=self.client, endpoint="tables/management/flow/netflow/devices", sn=self.sn)

    @property
    def netflow_collectors(self):
        return models.Table(client=self.client, endpoint="tables/management/flow/netflow/collectors", sn=self.sn)

    @property
    def netflow_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/management/flow/netflow/interfaces", sn=self.sn)

    @property
    def sflow_devices(self):
        return models.Table(client=self.client, endpoint="tables/management/flow/sflow/devices", sn=self.sn)

    @property
    def sflow_collectors(self):
        return models.Table(client=self.client, endpoint="tables/management/flow/sflow/collectors", sn=self.sn)

    @property
    def sflow_sources(self):
        return models.Table(client=self.client, endpoint="tables/management/flow/sflow/sources", sn=self.sn)

    @property
    def snmp_summary(self):
        return models.Table(client=self.client, endpoint="tables/management/snmp/summary", sn=self.sn)

    @property
    def snmp_communities(self):
        return models.Table(client=self.client, endpoint="tables/management/snmp/communities", sn=self.sn)

    @property
    def snmp_trap_hosts(self):
        return models.Table(client=self.client, endpoint="tables/management/snmp/trap-hosts", sn=self.sn)

    @property
    def snmp_users(self):
        return models.Table(client=self.client, endpoint="tables/management/snmp/users", sn=self.sn)

    @property
    def ptp_local_clock(self):
        return models.Table(client=self.client, endpoint="tables/management/ptp/local-clock", sn=self.sn)

    @property
    def ptp_masters(self):
        return models.Table(client=self.client, endpoint="tables/management/ptp/masters", sn=self.sn)

    @property
    def ptp_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/management/ptp/interfaces", sn=self.sn)

    @property
    def license_summary(self):
        return models.Table(client=self.client, endpoint="tables/management/licenses/summary", sn=self.sn)

    @property
    def licenses(self):
        return models.Table(client=self.client, endpoint="tables/management/licenses", sn=self.sn)

    @property
    def licenses_detail(self):
        return models.Table(client=self.client, endpoint="tables/management/licenses/detail", sn=self.sn)

    @property
    def cisco_smart_licenses_authorization(self):
        return models.Table(
            client=self.client, endpoint="tables/management/licenses/cisco-smart-licenses/authorization", sn=self.sn
        )

    @property
    def cisco_smart_licenses_registration(self):
        return models.Table(
            client=self.client, endpoint="tables/management/licenses/cisco-smart-licenses/registration", sn=self.sn
        )

    @property
    def cisco_smart_licenses_reservations(self):
        return models.Table(
            client=self.client, endpoint="tables/management/licenses/cisco-smart-licenses/reservations", sn=self.sn
        )

    @property
    def dns_resolver_settings(self):
        return models.Table(client=self.client, endpoint="tables/management/dns/settings", sn=self.sn)

    @property
    def dns_resolver_servers(self):
        return models.Table(client=self.client, endpoint="tables/management/dns/servers", sn=self.sn)

    @property
    def banners_summary(self):
        return models.Table(client=self.client, endpoint="tables/management/banners/summary", sn=self.sn)

    @property
    def banners_detail(self):
        return models.Table(client=self.client, endpoint="tables/management/banners/banners", sn=self.sn)



from typing import Union, List, Optional

from pydantic import BaseModel, Field


class Checks(BaseModel):
    """model for intent checks"""

    green: Union[int, str, dict] = Field(None, alias="0")
    blue: Union[int, str, dict] = Field(None, alias="10")
    amber: Union[int, str, dict] = Field(None, alias="20")
    red: Union[int, str, dict] = Field(None, alias="30")


class CheckResults(Checks):
    """model for intent check results"""

    nan: Union[int, str, dict] = Field(None, alias="-1")


class Description(BaseModel):
    """model for description of intent check"""

    general: Union[None, str] = None
    checks: Checks = Field(default_factory=Checks)


class Result(BaseModel):
    """model for results of intent check"""

    count: Union[int, None] = None
    checks: Checks = Field(default_factory=Checks)

    def compare(self, other: "Result") -> dict:
        """

        Args:
            other: intent check

        Returns:
            diction to use to compare intent checks
        """

        old = self.checks
        new = other.checks
        data = dict()
        if self.count is not None or other.count is not None:
            data["count"] = dict(
                loaded_snapshot=self.count or 0,
                compare_snapshot=other.count or 0,
                diff=(other.count or 0) - (self.count or 0),
            )

        for value in ["green", "blue", "amber", "red"]:
            if getattr(old, value) is not None or getattr(new, value) is not None:
                o = self.get_value(old, value)
                n = self.get_value(new, value)
                data[value] = dict(loaded_snapshot=o, compare_snapshot=n, diff=(n - o))
        return data

    @staticmethod
    def get_value(data: Checks, value: str):
        return int((getattr(data, value) if data else 0) or 0)


class Child(BaseModel):
    """model for child of intent check"""

    weight: int
    intent_id: str = Field(None, alias="id")


class Group(BaseModel):
    """model of a group of intent checks"""

    custom: bool
    name: str
    group_id: str = Field(None, alias="id")
    children: List[Child] = Field(default_factory=list)


class IntentCheck(BaseModel):
    """model for intent checks"""

    groups: List[Group]
    checks: Checks
    column: str
    custom: bool
    descriptions: Description
    name: str
    status: int
    result: Result
    api_endpoint: str = Field(None, alias="apiEndpoint")
    default_color: Union[None, int] = Field(None, alias="defaultColor")
    web_endpoint: str = Field(None, alias="webEndpoint")
    intent_id: str = Field(None, alias="id")
    result_data: Optional[CheckResults] = Field(default_factory=CheckResults)



import os
import re
from copy import deepcopy
from ipaddress import IPv4Interface
from typing import Optional, Union, List

try:
    from typing import Annotated  # py38 required Annotated
except ImportError:
    from typing_extensions import Annotated

from pydantic import field_validator, BaseModel, Field
from pydantic.functional_validators import BeforeValidator
from ipfabric.tools import validate_ip_network_str, VALID_IP

from .constants import VALID_LAYOUTS

PYDANTIC_EXTRAS = os.getenv("IPFABRIC_PYDANTIC_EXTRAS", "allow")
PORT_REGEX = re.compile(r"^\d*$|^\d*-\d*$")
ALL_NETWORK = "$main"

IPv4 = Annotated[Union[str, VALID_IP], BeforeValidator(validate_ip_network_str)]


class Instance(BaseModel, extra=PYDANTIC_EXTRAS):
    rootId: str
    vlanId: int
    visible: bool = True
    grouped: bool = True


class STPInstances(BaseModel, extra=PYDANTIC_EXTRAS):
    isolate: bool = False
    instances: List[Instance]


class Technologies(BaseModel, extra=PYDANTIC_EXTRAS):
    expandDeviceGroups: Optional[List[str]] = Field(default_factory=list)
    stpInstances: Optional[STPInstances] = None

    def technologies_parameters(self) -> dict:
        params = dict(expandDeviceGroups=self.expandDeviceGroups)
        if self.stpInstances:
            params["stpInstances"] = dict(isolate=self.stpInstances.isolate, instances=list())
            for i in self.stpInstances.instances:
                params["stpInstances"]["instances"].append(vars(i))
        return params


class ICMP(BaseModel, extra=PYDANTIC_EXTRAS):
    type: int
    code: int


class OtherOptions(BaseModel, extra=PYDANTIC_EXTRAS):
    applications: Optional[str] = ".*"
    tracked: Optional[bool] = False


class EntryPoint(BaseModel, extra=PYDANTIC_EXTRAS):
    sn: str = Field(title="Serial Number", description="IP Fabric Unique Device Serial Number API column sn")
    iface: str = Field(
        title="Interface", description="Interface to use as entry point. This is the intName not nameOriginal."
    )
    hostname: str = Field(title="Hostname", description="Hostname of the Device")


class Algorithm(BaseModel):
    """Default is automatic. Adding entryPoints will change to userDefined."""

    vrf: Optional[str] = None
    entryPoints: Optional[List[EntryPoint]] = None

    @property
    def type(self):
        return "userDefined" if self.entryPoints else "automatic"

    def algorithm_parameters(self) -> dict:
        if self.entryPoints:
            return dict(type=self.type, entryPoints=[vars(e) for e in self.entryPoints])
        else:
            return dict(type=self.type, vrf=self.vrf) if self.vrf else dict(type=self.type)


class PathLookup(BaseModel, extra=PYDANTIC_EXTRAS):
    protocol: Optional[str] = Field("tcp", title="Protocol", description="Valid protocols are tcp, udp, or icmp.")
    srcPorts: Optional[Union[str, int]] = Field(
        "1024-65535",
        title="Source Ports",
        description="Source ports if protocol is tcp or udp. "
        "Can be comma separated, a range using -, or any combination.",
    )
    dstPorts: Optional[Union[str, int]] = Field(
        "80,443",
        title="Destination Ports",
        description="Destination ports if protocol is tcp or udp. "
        "Can be comma separated, a range using -, or any combination.",
    )
    tcpFlags: Optional[list] = Field(
        default_factory=list,
        title="TCP Flags",
        description="Optional additional flags if protocol = TCP. "
        "Valid flags are ['ack', 'fin', 'psh', 'rst', 'syn', 'urg']",
    )
    icmp: Optional[ICMP] = Field(
        ICMP(type=0, code=0),
        title="ICMP Packet",
        description="Default is Echo Reply (type=0, code=0). You can pass in an ICMP model from ipfabric.diagrams.icmp "
        "or specify your own values like {'type': 1, 'code': 2}.",
    )
    ttl: Optional[int] = Field(128, title="Time To Live", description="TTL value, default is 128.")
    fragmentOffset: Optional[int] = Field(
        0, title="Fragment Offset", description="Fragment Offset value, default is 0."
    )
    securedPath: Optional[bool] = True
    enableRegions: Optional[bool] = False
    srcRegions: Optional[str] = ".*"
    dstRegions: Optional[str] = ".*"
    otherOptions: Optional[OtherOptions] = Field(default_factory=OtherOptions)
    firstHopAlgorithm: Optional[Algorithm] = Field(default_factory=Algorithm)

    @field_validator("protocol")
    @classmethod
    def _valid_protocols(cls, v):
        if v.lower() not in ["tcp", "udp", "icmp"]:
            raise ValueError(f'Protocol "{v}" not in ["tcp", "udp", "icmp"]')
        return v.lower()

    @field_validator("srcPorts", "dstPorts")
    @classmethod
    def _check_ports(cls, v):
        ports = str(v).replace(" ", "").split(",")
        for p in ports:
            if not PORT_REGEX.match(p):
                raise ValueError(
                    f'Ports "{v}" is not in the valid syntax, examples: ["80", "80,443", "0-1024", "80,8000-8100,8443"]'
                )
            if "-" in p:
                pn = p.split("-")
                if int(pn[0]) >= int(pn[1]):
                    raise ValueError(f'Ports "{p}" is invalid. {pn[0]} must be smaller than {pn[1]}.')
        return str(",".join(ports))

    @field_validator("tcpFlags")
    @classmethod
    def _valid_flags(cls, v):
        v = [f.lower() for f in v] if v else list()
        if all(f in ["ack", "fin", "psh", "rst", "syn", "urg"] for f in v):
            return v
        raise ValueError(f'TCP Flags "{v}" must be None or combination of ["ack", "fin", "psh", "rst", "syn", "urg"]')

    def _l4_options(self):
        if self.protocol == "icmp":
            return dict(type=self.icmp.type, code=self.icmp.code)
        elif self.protocol == "udp":
            return dict(srcPorts=self.srcPorts, dstPorts=self.dstPorts)
        else:
            return dict(srcPorts=self.srcPorts, dstPorts=self.dstPorts, flags=self.tcpFlags)

    def base_parameters(self) -> dict:
        return dict(
            type="pathLookup",
            groupBy="siteName",
            protocol=self.protocol,
            ttl=self.ttl,
            fragmentOffset=self.fragmentOffset,
            securedPath=self.securedPath,
            enableRegions=self.enableRegions,
            srcRegions=self.srcRegions,
            dstRegions=self.dstRegions,
            l4Options=self._l4_options(),
            otherOptions=vars(self.otherOptions),
            firstHopAlgorithm=self.firstHopAlgorithm.algorithm_parameters(),
        )

    @staticmethod
    def swap_src_dst(parameters: dict) -> dict:
        params = deepcopy(parameters)
        if params["protocol"] != "icmp":
            params["l4Options"]["srcPorts"], params["l4Options"]["dstPorts"] = (
                params["l4Options"]["dstPorts"],
                params["l4Options"]["srcPorts"],
            )
        params["srcRegions"], params["dstRegions"] = params["dstRegions"], params["srcRegions"]
        params["startingPoint"], params["destinationPoint"] = params["destinationPoint"], params["startingPoint"]
        return params


class Multicast(PathLookup, BaseModel, extra=PYDANTIC_EXTRAS):
    group: IPv4
    source: IPv4
    receiver: Optional[IPv4] = None

    @field_validator("group")
    @classmethod
    def _valid_group(cls, v):
        ip = IPv4Interface(v)
        if not ip.network.is_multicast:
            raise ValueError(f'Group IP "{v}" is not a valid Multicast Address.')
        return v

    def parameters(self) -> dict:
        parameters = self.base_parameters()
        parameters.update(
            dict(
                pathLookupType="multicast",
                group=str(self.group),
                source=str(self.source),
            )
        )
        if self.receiver:
            parameters["receiver"] = str(self.receiver)
        return parameters


class Unicast(PathLookup, BaseModel, extra=PYDANTIC_EXTRAS):
    startingPoint: IPv4 = Field(title="Source IP Address or Subnet")
    destinationPoint: IPv4 = Field(title="Destination IP Address or Subnet")

    def parameters(self, swap: bool = False) -> dict:
        parameters = self.base_parameters()
        parameters.update(
            dict(
                pathLookupType="unicast",
                networkMode=self._check_subnets(),
                startingPoint=self.startingPoint,
                destinationPoint=self.destinationPoint,
            )
        )
        return self.swap_src_dst(parameters) if swap else parameters

    def _check_subnets(self) -> bool:
        """
        Checks for valid IP Addresses or Subnet
        :param ips: ip addresses
        :return: True if a subnet is found to set to networkMode, False if only hosts
        """
        masks = {IPv4Interface(ip).network.prefixlen for ip in [self.startingPoint, self.destinationPoint]}
        return True if masks != {32} else False


class Host2GW(BaseModel, extra=PYDANTIC_EXTRAS):
    startingPoint: IPv4
    vrf: Optional[str] = None

    def parameters(self) -> dict:
        parameters = dict(
            pathLookupType="hostToDefaultGW",
            type="pathLookup",
            groupBy="siteName",
            startingPoint=self.startingPoint,
        )
        if self.vrf:
            parameters["vrf"] = self.vrf
        return parameters


class Layout(BaseModel, extra=PYDANTIC_EXTRAS):
    path: str
    layout: str

    @field_validator("layout")
    @classmethod
    def _valid_layout(cls, v):
        if v and v not in VALID_LAYOUTS:
            raise ValueError(f'Layout "{v}" is not in the valid layouts of {VALID_LAYOUTS}')
        return v


class Network(BaseModel, extra=PYDANTIC_EXTRAS):
    sites: Optional[Union[str, List[str]]] = [ALL_NETWORK]
    all_network: Optional[bool] = Field(False, description="Show all sites as clouds, UI option 'All Network'")
    layouts: Optional[List[Layout]] = None
    technologies: Optional[Technologies] = None

    @field_validator("sites")
    @classmethod
    def _format_paths(cls, v):
        if isinstance(v, str):
            return [v]
        return v

    def parameters(self) -> dict:
        parameters = dict(type="topology", groupBy="siteName", paths=self.sites.copy())
        if self.all_network and ALL_NETWORK not in parameters["paths"]:
            parameters["paths"].append(ALL_NETWORK)
        if self.layouts:
            parameters["layouts"] = [vars(layout) for layout in self.layouts]
        if self.technologies:
            parameters["technologies"] = self.technologies.technologies_parameters()
        return parameters



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Interfaces(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def current_rates_data_inbound(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/load/inbound", sn=self.sn)

    @property
    def current_rates_data_outbound(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/load/outbound", sn=self.sn)

    @property
    def current_rates_data_bidirectional(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/load/bidirectional", sn=self.sn)

    @property
    def average_rates_data_inbound(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/transfer-rates/inbound", sn=self.sn)

    @property
    def average_rates_data_outbound(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/transfer-rates/outbound", sn=self.sn)

    @property
    def average_rates_data_bidirectional(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/transfer-rates/bidirectional", sn=self.sn)

    @property
    def average_rates_data_inbound_per_device(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/transfer-rates/inbound-device", sn=self.sn)

    @property
    def average_rates_data_outbound_per_device(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/transfer-rates/outbound-device", sn=self.sn)

    @property
    def average_rates_data_bidirectional_per_device(self):
        return models.Table(
            client=self.client, endpoint="tables/interfaces/transfer-rates/bidirectional-device", sn=self.sn
        )

    @property
    def average_rates_errors_inbound(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/errors/inbound", sn=self.sn)

    @property
    def average_rates_errors_outbound(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/errors/outbound", sn=self.sn)

    @property
    def average_rates_errors_bidirectional(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/errors/bidirectional", sn=self.sn)

    @property
    def average_rates_errors_inbound_per_device(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/errors/inbound-device", sn=self.sn)

    @property
    def average_rates_errors_outbound_per_device(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/errors/outbound-device", sn=self.sn)

    @property
    def average_rates_errors_bidirectional_per_device(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/errors/bidirectional-device", sn=self.sn)

    @property
    def average_rates_drops_inbound(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/drops/inbound", sn=self.sn)

    @property
    def average_rates_drops_outbound(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/drops/outbound", sn=self.sn)

    @property
    def average_rates_drops_bidirectional(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/drops/bidirectional", sn=self.sn)

    @property
    def average_rates_drops_inbound_per_device(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/drops/inbound-device", sn=self.sn)

    @property
    def average_rates_drops_outbound_per_device(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/drops/outbound-device", sn=self.sn)

    @property
    def average_rates_drops_bidirectional_per_device(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/drops/bidirectional-device", sn=self.sn)

    @property
    def duplex(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/duplex", sn=self.sn)

    @property
    def err_disabled(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/errors/disabled", sn=self.sn)

    @property
    def connectivity_matrix(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/connectivity-matrix", sn=self.sn)

    @property
    def connectivity_matrix_unmanaged_neighbors_summary(self):
        return models.Table(
            client=self.client, endpoint="tables/interfaces/connectivity-matrix/unmanaged-neighbors/summary", sn=self.sn
        )

    @property
    def connectivity_matrix_unmanaged_neighbors_detail(self):
        return models.Table(
            client=self.client, endpoint="tables/interfaces/connectivity-matrix/unmanaged-neighbors/detail", sn=self.sn
        )

    @property
    def switchport(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/switchports", sn=self.sn)

    @property
    def mtu(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/mtu", sn=self.sn)

    @property
    def storm_control_all(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/storm-control/all", sn=self.sn)

    @property
    def storm_control_broadcast(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/storm-control/broadcast", sn=self.sn)

    @property
    def storm_control_unicast(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/storm-control/unicast", sn=self.sn)

    @property
    def storm_control_multicast(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/storm-control/multicast", sn=self.sn)

    @property
    def transceivers(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/transceivers/inventory", sn=self.sn)

    @property
    def transceivers_statistics(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/transceivers/statistics", sn=self.sn)

    @property
    def transceivers_triggered_thresholds(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/transceivers/thresholds", sn=self.sn)

    @property
    def transceivers_errors(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/transceivers/errors", sn=self.sn)

    @property
    def point_to_point_over_ethernet(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/pppoe", sn=self.sn)

    @property
    def point_to_point_over_ethernet_sessions(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/pppoe/sessions", sn=self.sn)

    @property
    def counters_inbound(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/counters/inbound", sn=self.sn)

    @property
    def counters_outbound(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/counters/outbound", sn=self.sn)

    @property
    def tunnels_ipv4(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/tunnels/ipv4", sn=self.sn)

    @property
    def tunnels_ipv6(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/tunnels/ipv6", sn=self.sn)



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Routing(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def summary_protocols(self):
        return models.Table(client=self.client, endpoint="tables/networks/summary/protocols", sn=self.sn)

    @property
    def summary_protocols_bgp(self):
        return models.Table(client=self.client, endpoint="tables/networks/summary/protocols/bgp", sn=self.sn)

    @property
    def summary_protocols_eigrp(self):
        return models.Table(client=self.client, endpoint="tables/networks/summary/protocols/eigrp", sn=self.sn)

    @property
    def summary_protocols_isis(self):
        return models.Table(client=self.client, endpoint="tables/networks/summary/protocols/isis", sn=self.sn)

    @property
    def summary_protocols_ospf(self):
        return models.Table(client=self.client, endpoint="tables/networks/summary/protocols/ospf", sn=self.sn)

    @property
    def summary_protocols_ospfv3(self):
        return models.Table(client=self.client, endpoint="tables/networks/summary/protocols/ospfv3", sn=self.sn)

    @property
    def summary_protocols_rip(self):
        return models.Table(client=self.client, endpoint="tables/networks/summary/protocols/rip", sn=self.sn)

    @property
    def routes_ipv4(self):
        return models.Table(client=self.client, endpoint="tables/networks/routes", sn=self.sn)

    @property
    def routes_ipv6(self):
        return models.Table(client=self.client, endpoint="tables/networks/ipv6-routes", sn=self.sn)

    @property
    def route_stability(self):
        return models.Table(client=self.client, endpoint="tables/networks/route-stability", sn=self.sn)

    @property
    def ospf_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/ospf/neighbors", sn=self.sn)

    @property
    def ospf_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/ospf/interfaces", sn=self.sn)

    @property
    def ospfv3_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/ospf-v3/neighbors", sn=self.sn)

    @property
    def ospfv3_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/ospf-v3/interfaces", sn=self.sn)

    @property
    def bgp_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/bgp/neighbors", sn=self.sn)

    @property
    def bgp_address_families(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/bgp/address-families", sn=self.sn)

    @property
    def eigrp_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/eigrp/neighbors", sn=self.sn)

    @property
    def eigrp_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/eigrp/interfaces", sn=self.sn)

    @property
    def rip_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/rip/neighbors", sn=self.sn)

    @property
    def rip_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/rip/interfaces", sn=self.sn)

    @property
    def isis_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/is-is/neighbors", sn=self.sn)

    @property
    def isis_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/is-is/interfaces", sn=self.sn)

    @property
    def isis_levels(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/is-is/levels", sn=self.sn)

    @property
    def path_lookup_checks(self):
        return models.Table(client=self.client, endpoint="tables/networks/path-lookup-checks", sn=self.sn)

    @property
    def vrf_summary(self):
        return models.Table(client=self.client, endpoint="tables/vrf/summary", sn=self.sn)

    @property
    def vrf_detail(self):
        return models.Table(client=self.client, endpoint="tables/vrf/detail", sn=self.sn)

    @property
    def vrf_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/vrf/interfaces", sn=self.sn)

    @property
    def policies_prefix_list(self):
        return models.Table(client=self.client, endpoint="tables/networks/policies/prefix-lists/ipv4", sn=self.sn)

    @property
    def policies_prefix_list_ipv6(self):
        return models.Table(client=self.client, endpoint="tables/networks/policies/prefix-lists/ipv6", sn=self.sn)

    @property
    def lisp_routes_ipv4(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/lisp/ipv4-routes", sn=self.sn)

    @property
    def lisp_routes_ipv6(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/lisp/ipv6-routes", sn=self.sn)

    @property
    def lisp_map_resolvers_ipv4(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/lisp/ipv4-databases", sn=self.sn)

    @property
    def lisp_map_resolvers_ipv6(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/lisp/ipv6-databases", sn=self.sn)

    @property
    def policies(self):
        return models.Table(client=self.client, endpoint="tables/networks/policies/routing", sn=self.sn)

    @property
    def policies_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/networks/policies/routing/interfaces", sn=self.sn)

    @property
    def policies_pbr(self):
        return models.Table(client=self.client, endpoint="tables/security/pbr", sn=self.sn)



import logging
from datetime import datetime, timezone
from ipaddress import IPv4Address, IPv4Network, IPv4Interface, AddressValueError
from typing import Union

import macaddress
from dateutil import parser
from pytz import BaseTzInfo, utc

logger = logging.getLogger("ipfabric")

LAST_ID, PREV_ID, LASTLOCKED_ID = "$last", "$prev", "$lastLocked"
VALID_REFS = [LAST_ID, PREV_ID, LASTLOCKED_ID]
VALID_IP = Union[IPv4Address, IPv4Network, IPv4Interface]


def validate_ip_network(ip: Union[str, VALID_IP]) -> IPv4Network:
    if isinstance(ip, IPv4Network):
        return ip
    elif isinstance(ip, (IPv4Address, IPv4Interface)):
        return IPv4Network(ip, strict=False)
    else:
        try:
            return IPv4Network(ip, strict=False)
        except AddressValueError:
            raise ValueError(f'IP "{ip}" not a valid IP Address')


def validate_ip_network_str(ip: Union[str, IPv4Address, IPv4Network, IPv4Interface]):
    return str((validate_ip_network(ip)))


def create_filter(client, url, filters=None, sn=None) -> tuple:
    all_columns = client.get_columns(url)
    if not any([_ in ["sn", "localSn"] for _ in all_columns]):
        raise KeyError(f'Column "sn" or "localSn" not found in "{url}" table.')
    f = {"sn": ["eq", sn]} if "sn" in all_columns else {"localSn": ["eq", sn]}
    if filters:
        f = {"and": [f, filters]}
    return all_columns, f


class MAC(macaddress.MAC):
    formats = ("xxxx.xxxx.xxxx",) + macaddress.MAC.formats


def convert_timestamp(
    ts: int, ts_format: str = "datetime", milliseconds: bool = True, tzinfo: BaseTzInfo = utc
) -> datetime:
    """Converts IP Fabric timestamp in milliseconds to a formatted time

    Args:
        ts: int: Timestamp in milliseconds (IP Fabric's returned format)
        ts_format: str: Default datetime, can be ['datetime', 'iso', 'ziso', 'date']
        milliseconds: bool: Default True to return milliseconds, False returns seconds
        tzinfo: pytz object: Default utc
    Returns:
        date time object in the requested format
    """
    dt = datetime.fromtimestamp(ts / 1000, tz=timezone.utc).astimezone(tzinfo)
    if ts_format == "datetime":
        return dt
    elif ts_format == "iso":
        return dt.isoformat(timespec="milliseconds" if milliseconds else "seconds")
    elif ts_format == "ziso":
        return dt.isoformat(timespec="milliseconds" if milliseconds else "seconds").replace("+00:00", "Z")
    elif ts_format == "date":
        return dt.date()
    else:
        raise SyntaxError(f"{ts_format} is not one of ['datetime', 'iso', 'ziso', 'date'].")


def date_parser(timestamp: Union[int, str]) -> datetime:
    return (
        datetime.fromtimestamp(timestamp, tz=timezone.utc)
        if isinstance(timestamp, int)
        else parser.parse(timestamp).replace(tzinfo=timezone.utc)
    )


def parse_mac(mac_address) -> Union[str, list, None]:
    """Takes any format MAC address and return IP Fabric format.

    Args:
        mac_address: Union[str, list]: String or list of MAC address
    Returns:
        String or list of MAC address
    """

    def mac_format(mac):
        try:
            return str(MAC(mac.strip())).lower()
        except ValueError:
            logger.warning(f"{mac.strip()} is not a valid mac.")
            return None

    if isinstance(mac_address, str):
        return mac_format(mac_address)
    elif isinstance(mac_address, list):
        parsed = list()
        for m in mac_address:
            if not m:
                continue
            mac = mac_format(m)
            if mac:
                parsed.append(mac)

        return parsed
    return None


TIMEZONES = {
    "africa/abidjan": "Africa/Abidjan",
    "africa/accra": "Africa/Accra",
    "africa/addis_ababa": "Africa/Addis_Ababa",
    "africa/algiers": "Africa/Algiers",
    "africa/asmara": "Africa/Asmara",
    "africa/bamako": "Africa/Bamako",
    "africa/bangui": "Africa/Bangui",
    "africa/banjul": "Africa/Banjul",
    "africa/bissau": "Africa/Bissau",
    "africa/blantyre": "Africa/Blantyre",
    "africa/brazzaville": "Africa/Brazzaville",
    "africa/bujumbura": "Africa/Bujumbura",
    "africa/cairo": "Africa/Cairo",
    "africa/casablanca": "Africa/Casablanca",
    "africa/ceuta": "Africa/Ceuta",
    "africa/conakry": "Africa/Conakry",
    "africa/dakar": "Africa/Dakar",
    "africa/dar_es_salaam": "Africa/Dar_es_Salaam",
    "africa/djibouti": "Africa/Djibouti",
    "africa/douala": "Africa/Douala",
    "africa/el_aaiun": "Africa/El_Aaiun",
    "africa/freetown": "Africa/Freetown",
    "africa/gaborone": "Africa/Gaborone",
    "africa/harare": "Africa/Harare",
    "africa/johannesburg": "Africa/Johannesburg",
    "africa/juba": "Africa/Juba",
    "africa/kampala": "Africa/Kampala",
    "africa/khartoum": "Africa/Khartoum",
    "africa/kigali": "Africa/Kigali",
    "africa/kinshasa": "Africa/Kinshasa",
    "africa/lagos": "Africa/Lagos",
    "africa/libreville": "Africa/Libreville",
    "africa/lome": "Africa/Lome",
    "africa/luanda": "Africa/Luanda",
    "africa/lubumbashi": "Africa/Lubumbashi",
    "africa/lusaka": "Africa/Lusaka",
    "africa/malabo": "Africa/Malabo",
    "africa/maputo": "Africa/Maputo",
    "africa/maseru": "Africa/Maseru",
    "africa/mbabane": "Africa/Mbabane",
    "africa/mogadishu": "Africa/Mogadishu",
    "africa/monrovia": "Africa/Monrovia",
    "africa/nairobi": "Africa/Nairobi",
    "africa/ndjamena": "Africa/Ndjamena",
    "africa/niamey": "Africa/Niamey",
    "africa/nouakchott": "Africa/Nouakchott",
    "africa/ouagadougou": "Africa/Ouagadougou",
    "africa/porto-novo": "Africa/Porto-Novo",
    "africa/sao_tome": "Africa/Sao_Tome",
    "africa/tripoli": "Africa/Tripoli",
    "africa/tunis": "Africa/Tunis",
    "africa/windhoek": "Africa/Windhoek",
    "america/adak": "America/Adak",
    "america/anchorage": "America/Anchorage",
    "america/anguilla": "America/Anguilla",
    "america/antigua": "America/Antigua",
    "america/araguaina": "America/Araguaina",
    "america/argentina/buenos_aires": "America/Argentina/Buenos_Aires",
    "america/argentina/catamarca": "America/Argentina/Catamarca",
    "america/argentina/cordoba": "America/Argentina/Cordoba",
    "america/argentina/jujuy": "America/Argentina/Jujuy",
    "america/argentina/la_rioja": "America/Argentina/La_Rioja",
    "america/argentina/mendoza": "America/Argentina/Mendoza",
    "america/argentina/rio_gallegos": "America/Argentina/Rio_Gallegos",
    "america/argentina/salta": "America/Argentina/Salta",
    "america/argentina/san_juan": "America/Argentina/San_Juan",
    "america/argentina/san_luis": "America/Argentina/San_Luis",
    "america/argentina/tucuman": "America/Argentina/Tucuman",
    "america/argentina/ushuaia": "America/Argentina/Ushuaia",
    "america/aruba": "America/Aruba",
    "america/asuncion": "America/Asuncion",
    "america/atikokan": "America/Atikokan",
    "america/bahia": "America/Bahia",
    "america/bahia_banderas": "America/Bahia_Banderas",
    "america/barbados": "America/Barbados",
    "america/belem": "America/Belem",
    "america/belize": "America/Belize",
    "america/blanc-sablon": "America/Blanc-Sablon",
    "america/boa_vista": "America/Boa_Vista",
    "america/bogota": "America/Bogota",
    "america/boise": "America/Boise",
    "america/cambridge_bay": "America/Cambridge_Bay",
    "america/campo_grande": "America/Campo_Grande",
    "america/cancun": "America/Cancun",
    "america/caracas": "America/Caracas",
    "america/cayenne": "America/Cayenne",
    "america/cayman": "America/Cayman",
    "america/chicago": "America/Chicago",
    "america/chihuahua": "America/Chihuahua",
    "america/costa_rica": "America/Costa_Rica",
    "america/creston": "America/Creston",
    "america/cuiaba": "America/Cuiaba",
    "america/curacao": "America/Curacao",
    "america/danmarkshavn": "America/Danmarkshavn",
    "america/dawson": "America/Dawson",
    "america/dawson_creek": "America/Dawson_Creek",
    "america/denver": "America/Denver",
    "america/detroit": "America/Detroit",
    "america/dominica": "America/Dominica",
    "america/edmonton": "America/Edmonton",
    "america/eirunepe": "America/Eirunepe",
    "america/el_salvador": "America/El_Salvador",
    "america/fort_nelson": "America/Fort_Nelson",
    "america/fortaleza": "America/Fortaleza",
    "america/glace_bay": "America/Glace_Bay",
    "america/goose_bay": "America/Goose_Bay",
    "america/grand_turk": "America/Grand_Turk",
    "america/grenada": "America/Grenada",
    "america/guadeloupe": "America/Guadeloupe",
    "america/guatemala": "America/Guatemala",
    "america/guayaquil": "America/Guayaquil",
    "america/guyana": "America/Guyana",
    "america/halifax": "America/Halifax",
    "america/havana": "America/Havana",
    "america/hermosillo": "America/Hermosillo",
    "america/indiana/indianapolis": "America/Indiana/Indianapolis",
    "america/indiana/knox": "America/Indiana/Knox",
    "america/indiana/marengo": "America/Indiana/Marengo",
    "america/indiana/petersburg": "America/Indiana/Petersburg",
    "america/indiana/tell_city": "America/Indiana/Tell_City",
    "america/indiana/vevay": "America/Indiana/Vevay",
    "america/indiana/vincennes": "America/Indiana/Vincennes",
    "america/indiana/winamac": "America/Indiana/Winamac",
    "america/inuvik": "America/Inuvik",
    "america/iqaluit": "America/Iqaluit",
    "america/jamaica": "America/Jamaica",
    "america/juneau": "America/Juneau",
    "america/kentucky/louisville": "America/Kentucky/Louisville",
    "america/kentucky/monticello": "America/Kentucky/Monticello",
    "america/kralendijk": "America/Kralendijk",
    "america/la_paz": "America/La_Paz",
    "america/lima": "America/Lima",
    "america/los_angeles": "America/Los_Angeles",
    "america/lower_princes": "America/Lower_Princes",
    "america/maceio": "America/Maceio",
    "america/managua": "America/Managua",
    "america/manaus": "America/Manaus",
    "america/marigot": "America/Marigot",
    "america/martinique": "America/Martinique",
    "america/matamoros": "America/Matamoros",
    "america/mazatlan": "America/Mazatlan",
    "america/menominee": "America/Menominee",
    "america/merida": "America/Merida",
    "america/metlakatla": "America/Metlakatla",
    "america/mexico_city": "America/Mexico_City",
    "america/miquelon": "America/Miquelon",
    "america/moncton": "America/Moncton",
    "america/monterrey": "America/Monterrey",
    "america/montevideo": "America/Montevideo",
    "america/montserrat": "America/Montserrat",
    "america/nassau": "America/Nassau",
    "america/new_york": "America/New_York",
    "america/nipigon": "America/Nipigon",
    "america/nome": "America/Nome",
    "america/noronha": "America/Noronha",
    "america/north_dakota/beulah": "America/North_Dakota/Beulah",
    "america/north_dakota/center": "America/North_Dakota/Center",
    "america/north_dakota/new_salem": "America/North_Dakota/New_Salem",
    "america/nuuk": "America/Nuuk",
    "america/ojinaga": "America/Ojinaga",
    "america/panama": "America/Panama",
    "america/pangnirtung": "America/Pangnirtung",
    "america/paramaribo": "America/Paramaribo",
    "america/phoenix": "America/Phoenix",
    "america/port-au-prince": "America/Port-au-Prince",
    "america/port_of_spain": "America/Port_of_Spain",
    "america/porto_velho": "America/Porto_Velho",
    "america/puerto_rico": "America/Puerto_Rico",
    "america/punta_arenas": "America/Punta_Arenas",
    "america/rainy_river": "America/Rainy_River",
    "america/rankin_inlet": "America/Rankin_Inlet",
    "america/recife": "America/Recife",
    "america/regina": "America/Regina",
    "america/resolute": "America/Resolute",
    "america/rio_branco": "America/Rio_Branco",
    "america/santarem": "America/Santarem",
    "america/santiago": "America/Santiago",
    "america/santo_domingo": "America/Santo_Domingo",
    "america/sao_paulo": "America/Sao_Paulo",
    "america/scoresbysund": "America/Scoresbysund",
    "america/sitka": "America/Sitka",
    "america/st_barthelemy": "America/St_Barthelemy",
    "america/st_johns": "America/St_Johns",
    "america/st_kitts": "America/St_Kitts",
    "america/st_lucia": "America/St_Lucia",
    "america/st_thomas": "America/St_Thomas",
    "america/st_vincent": "America/St_Vincent",
    "america/swift_current": "America/Swift_Current",
    "america/tegucigalpa": "America/Tegucigalpa",
    "america/thule": "America/Thule",
    "america/thunder_bay": "America/Thunder_Bay",
    "america/tijuana": "America/Tijuana",
    "america/toronto": "America/Toronto",
    "america/tortola": "America/Tortola",
    "america/vancouver": "America/Vancouver",
    "america/whitehorse": "America/Whitehorse",
    "america/winnipeg": "America/Winnipeg",
    "america/yakutat": "America/Yakutat",
    "america/yellowknife": "America/Yellowknife",
    "antarctica/casey": "Antarctica/Casey",
    "antarctica/davis": "Antarctica/Davis",
    "antarctica/dumontdurville": "Antarctica/DumontDUrville",
    "antarctica/macquarie": "Antarctica/Macquarie",
    "antarctica/mawson": "Antarctica/Mawson",
    "antarctica/mcmurdo": "Antarctica/McMurdo",
    "antarctica/palmer": "Antarctica/Palmer",
    "antarctica/rothera": "Antarctica/Rothera",
    "antarctica/syowa": "Antarctica/Syowa",
    "antarctica/troll": "Antarctica/Troll",
    "antarctica/vostok": "Antarctica/Vostok",
    "arctic/longyearbyen": "Arctic/Longyearbyen",
    "asia/aden": "Asia/Aden",
    "asia/almaty": "Asia/Almaty",
    "asia/amman": "Asia/Amman",
    "asia/anadyr": "Asia/Anadyr",
    "asia/aqtau": "Asia/Aqtau",
    "asia/aqtobe": "Asia/Aqtobe",
    "asia/ashgabat": "Asia/Ashgabat",
    "asia/atyrau": "Asia/Atyrau",
    "asia/baghdad": "Asia/Baghdad",
    "asia/bahrain": "Asia/Bahrain",
    "asia/baku": "Asia/Baku",
    "asia/bangkok": "Asia/Bangkok",
    "asia/barnaul": "Asia/Barnaul",
    "asia/beirut": "Asia/Beirut",
    "asia/bishkek": "Asia/Bishkek",
    "asia/brunei": "Asia/Brunei",
    "asia/chita": "Asia/Chita",
    "asia/choibalsan": "Asia/Choibalsan",
    "asia/colombo": "Asia/Colombo",
    "asia/damascus": "Asia/Damascus",
    "asia/dhaka": "Asia/Dhaka",
    "asia/dili": "Asia/Dili",
    "asia/dubai": "Asia/Dubai",
    "asia/dushanbe": "Asia/Dushanbe",
    "asia/famagusta": "Asia/Famagusta",
    "asia/gaza": "Asia/Gaza",
    "asia/hebron": "Asia/Hebron",
    "asia/ho_chi_minh": "Asia/Ho_Chi_Minh",
    "asia/hong_kong": "Asia/Hong_Kong",
    "asia/hovd": "Asia/Hovd",
    "asia/irkutsk": "Asia/Irkutsk",
    "asia/jakarta": "Asia/Jakarta",
    "asia/jayapura": "Asia/Jayapura",
    "asia/jerusalem": "Asia/Jerusalem",
    "asia/kabul": "Asia/Kabul",
    "asia/kamchatka": "Asia/Kamchatka",
    "asia/karachi": "Asia/Karachi",
    "asia/kathmandu": "Asia/Kathmandu",
    "asia/khandyga": "Asia/Khandyga",
    "asia/kolkata": "Asia/Kolkata",
    "asia/krasnoyarsk": "Asia/Krasnoyarsk",
    "asia/kuala_lumpur": "Asia/Kuala_Lumpur",
    "asia/kuching": "Asia/Kuching",
    "asia/kuwait": "Asia/Kuwait",
    "asia/macau": "Asia/Macau",
    "asia/magadan": "Asia/Magadan",
    "asia/makassar": "Asia/Makassar",
    "asia/manila": "Asia/Manila",
    "asia/muscat": "Asia/Muscat",
    "asia/nicosia": "Asia/Nicosia",
    "asia/novokuznetsk": "Asia/Novokuznetsk",
    "asia/novosibirsk": "Asia/Novosibirsk",
    "asia/omsk": "Asia/Omsk",
    "asia/oral": "Asia/Oral",
    "asia/phnom_penh": "Asia/Phnom_Penh",
    "asia/pontianak": "Asia/Pontianak",
    "asia/pyongyang": "Asia/Pyongyang",
    "asia/qatar": "Asia/Qatar",
    "asia/qostanay": "Asia/Qostanay",
    "asia/qyzylorda": "Asia/Qyzylorda",
    "asia/riyadh": "Asia/Riyadh",
    "asia/sakhalin": "Asia/Sakhalin",
    "asia/samarkand": "Asia/Samarkand",
    "asia/seoul": "Asia/Seoul",
    "asia/shanghai": "Asia/Shanghai",
    "asia/singapore": "Asia/Singapore",
    "asia/srednekolymsk": "Asia/Srednekolymsk",
    "asia/taipei": "Asia/Taipei",
    "asia/tashkent": "Asia/Tashkent",
    "asia/tbilisi": "Asia/Tbilisi",
    "asia/tehran": "Asia/Tehran",
    "asia/thimphu": "Asia/Thimphu",
    "asia/tokyo": "Asia/Tokyo",
    "asia/tomsk": "Asia/Tomsk",
    "asia/ulaanbaatar": "Asia/Ulaanbaatar",
    "asia/urumqi": "Asia/Urumqi",
    "asia/ust-nera": "Asia/Ust-Nera",
    "asia/vientiane": "Asia/Vientiane",
    "asia/vladivostok": "Asia/Vladivostok",
    "asia/yakutsk": "Asia/Yakutsk",
    "asia/yangon": "Asia/Yangon",
    "asia/yekaterinburg": "Asia/Yekaterinburg",
    "asia/yerevan": "Asia/Yerevan",
    "atlantic/azores": "Atlantic/Azores",
    "atlantic/bermuda": "Atlantic/Bermuda",
    "atlantic/canary": "Atlantic/Canary",
    "atlantic/cape_verde": "Atlantic/Cape_Verde",
    "atlantic/faroe": "Atlantic/Faroe",
    "atlantic/madeira": "Atlantic/Madeira",
    "atlantic/reykjavik": "Atlantic/Reykjavik",
    "atlantic/south_georgia": "Atlantic/South_Georgia",
    "atlantic/st_helena": "Atlantic/St_Helena",
    "atlantic/stanley": "Atlantic/Stanley",
    "australia/adelaide": "Australia/Adelaide",
    "australia/brisbane": "Australia/Brisbane",
    "australia/broken_hill": "Australia/Broken_Hill",
    "australia/darwin": "Australia/Darwin",
    "australia/eucla": "Australia/Eucla",
    "australia/hobart": "Australia/Hobart",
    "australia/lindeman": "Australia/Lindeman",
    "australia/lord_howe": "Australia/Lord_Howe",
    "australia/melbourne": "Australia/Melbourne",
    "australia/perth": "Australia/Perth",
    "australia/sydney": "Australia/Sydney",
    "europe/amsterdam": "Europe/Amsterdam",
    "europe/andorra": "Europe/Andorra",
    "europe/astrakhan": "Europe/Astrakhan",
    "europe/athens": "Europe/Athens",
    "europe/belgrade": "Europe/Belgrade",
    "europe/berlin": "Europe/Berlin",
    "europe/bratislava": "Europe/Bratislava",
    "europe/brussels": "Europe/Brussels",
    "europe/bucharest": "Europe/Bucharest",
    "europe/budapest": "Europe/Budapest",
    "europe/busingen": "Europe/Busingen",
    "europe/chisinau": "Europe/Chisinau",
    "europe/copenhagen": "Europe/Copenhagen",
    "europe/dublin": "Europe/Dublin",
    "europe/gibraltar": "Europe/Gibraltar",
    "europe/guernsey": "Europe/Guernsey",
    "europe/helsinki": "Europe/Helsinki",
    "europe/isle_of_man": "Europe/Isle_of_Man",
    "europe/istanbul": "Europe/Istanbul",
    "europe/jersey": "Europe/Jersey",
    "europe/kaliningrad": "Europe/Kaliningrad",
    "europe/kiev": "Europe/Kiev",
    "europe/kirov": "Europe/Kirov",
    "europe/lisbon": "Europe/Lisbon",
    "europe/ljubljana": "Europe/Ljubljana",
    "europe/london": "Europe/London",
    "europe/luxembourg": "Europe/Luxembourg",
    "europe/madrid": "Europe/Madrid",
    "europe/malta": "Europe/Malta",
    "europe/mariehamn": "Europe/Mariehamn",
    "europe/minsk": "Europe/Minsk",
    "europe/monaco": "Europe/Monaco",
    "europe/moscow": "Europe/Moscow",
    "europe/oslo": "Europe/Oslo",
    "europe/paris": "Europe/Paris",
    "europe/podgorica": "Europe/Podgorica",
    "europe/prague": "Europe/Prague",
    "europe/riga": "Europe/Riga",
    "europe/rome": "Europe/Rome",
    "europe/samara": "Europe/Samara",
    "europe/san_marino": "Europe/San_Marino",
    "europe/sarajevo": "Europe/Sarajevo",
    "europe/saratov": "Europe/Saratov",
    "europe/simferopol": "Europe/Simferopol",
    "europe/skopje": "Europe/Skopje",
    "europe/sofia": "Europe/Sofia",
    "europe/stockholm": "Europe/Stockholm",
    "europe/tallinn": "Europe/Tallinn",
    "europe/tirane": "Europe/Tirane",
    "europe/ulyanovsk": "Europe/Ulyanovsk",
    "europe/uzhgorod": "Europe/Uzhgorod",
    "europe/vaduz": "Europe/Vaduz",
    "europe/vatican": "Europe/Vatican",
    "europe/vienna": "Europe/Vienna",
    "europe/vilnius": "Europe/Vilnius",
    "europe/volgograd": "Europe/Volgograd",
    "europe/warsaw": "Europe/Warsaw",
    "europe/zagreb": "Europe/Zagreb",
    "europe/zaporozhye": "Europe/Zaporozhye",
    "europe/zurich": "Europe/Zurich",
    "indian/antananarivo": "Indian/Antananarivo",
    "indian/chagos": "Indian/Chagos",
    "indian/christmas": "Indian/Christmas",
    "indian/cocos": "Indian/Cocos",
    "indian/comoro": "Indian/Comoro",
    "indian/kerguelen": "Indian/Kerguelen",
    "indian/mahe": "Indian/Mahe",
    "indian/maldives": "Indian/Maldives",
    "indian/mauritius": "Indian/Mauritius",
    "indian/mayotte": "Indian/Mayotte",
    "indian/reunion": "Indian/Reunion",
    "pacific/apia": "Pacific/Apia",
    "pacific/auckland": "Pacific/Auckland",
    "pacific/bougainville": "Pacific/Bougainville",
    "pacific/chatham": "Pacific/Chatham",
    "pacific/chuuk": "Pacific/Chuuk",
    "pacific/easter": "Pacific/Easter",
    "pacific/efate": "Pacific/Efate",
    "pacific/enderbury": "Pacific/Enderbury",
    "pacific/fakaofo": "Pacific/Fakaofo",
    "pacific/fiji": "Pacific/Fiji",
    "pacific/funafuti": "Pacific/Funafuti",
    "pacific/galapagos": "Pacific/Galapagos",
    "pacific/gambier": "Pacific/Gambier",
    "pacific/guadalcanal": "Pacific/Guadalcanal",
    "pacific/guam": "Pacific/Guam",
    "pacific/honolulu": "Pacific/Honolulu",
    "pacific/kiritimati": "Pacific/Kiritimati",
    "pacific/kosrae": "Pacific/Kosrae",
    "pacific/kwajalein": "Pacific/Kwajalein",
    "pacific/majuro": "Pacific/Majuro",
    "pacific/marquesas": "Pacific/Marquesas",
    "pacific/midway": "Pacific/Midway",
    "pacific/nauru": "Pacific/Nauru",
    "pacific/niue": "Pacific/Niue",
    "pacific/norfolk": "Pacific/Norfolk",
    "pacific/noumea": "Pacific/Noumea",
    "pacific/pago_pago": "Pacific/Pago_Pago",
    "pacific/palau": "Pacific/Palau",
    "pacific/pitcairn": "Pacific/Pitcairn",
    "pacific/pohnpei": "Pacific/Pohnpei",
    "pacific/port_moresby": "Pacific/Port_Moresby",
    "pacific/rarotonga": "Pacific/Rarotonga",
    "pacific/saipan": "Pacific/Saipan",
    "pacific/tahiti": "Pacific/Tahiti",
    "pacific/tarawa": "Pacific/Tarawa",
    "pacific/tongatapu": "Pacific/Tongatapu",
    "pacific/wake": "Pacific/Wake",
    "pacific/wallis": "Pacific/Wallis",
    "utc": "UTC",
}



import logging
from collections import defaultdict
from typing import Any, Union, List, Dict
from warnings import warn

from httpx import HTTPStatusError

from .intent_check import Group, IntentCheck

logger = logging.getLogger("ipfabric")
COLOR_DICT = dict(nan=-1, green=0, blue=10, amber=20, red=30)


def loaded_status(func):
    def _decorator(self, *args, **kwargs):
        if not self.loaded:
            logger.error("Intents are not loaded please run `ipf.intent.load_intent()`.")
        return func(self, *args, **kwargs)

    return _decorator


class Intent:
    def __init__(self, client, load=True):
        self.client: Any = client
        self.snapshot_id: str = self.client.snapshot_id
        self.loaded = self.snapshot_id and load
        self.intent_checks: List[IntentCheck] = self.get_intent_checks() if self.loaded else list()
        self.groups: List[Group] = self.get_groups() if self.loaded else list()

    def get_intent_checks(self, snapshot_id: str = None) -> List[IntentCheck]:
        """Gets all intent checks and returns a list of them.

        Args:
        snapshot_id: Optional snapshot ID to get different data

        Returns:
            list: List of intent checks
        """
        snapshot = self.client.snapshots[snapshot_id] if snapshot_id else self.client.snapshots[self.snapshot_id]
        if not snapshot.loaded:
            raise ValueError(f"Snapshot {snapshot.snapshot_id} is not loaded; cannot pull Intent Rules.")
        if snapshot.disabled_intent_verification is True:
            raise ValueError(
                f"Snapshot {snapshot.snapshot_id} has Intent Verification computation disabled; "
                "cannot pull Intent Rules."
            )
        res = self.client.get("reports", params=dict(snapshot=snapshot.snapshot_id))
        try:
            res.raise_for_status()
            return [IntentCheck(**check) for check in res.json()]
        except HTTPStatusError:
            logger.warning(self.client._api_insuf_rights + 'on GET "/reports". Will not load Intents.')
            return list()

    def load_intent(self, snapshot_id: str = None):
        """Loads intent checks into the class.

        Args:
            snapshot_id: Uses a different Snapshot ID then client
        """
        self.snapshot_id = snapshot_id or self.client.snapshot_id
        self.intent_checks = self.get_intent_checks(snapshot_id)
        self.groups = self.get_groups()
        self.loaded = True

    def get_groups(self) -> list:
        """

        Returns:
            list: list of groups
        """
        res = self.client.get("reports/groups")
        try:
            res.raise_for_status()
            return [Group(**group) for group in res.json()]
        except HTTPStatusError:
            logger.warning(self.client._api_insuf_rights + 'on GET "/reports/groups". Will not load Intent Groups.')
            return list()

    @property
    @loaded_status
    def custom(self) -> List[IntentCheck]:
        return [c for c in self.intent_checks if c.custom]

    @property
    @loaded_status
    def builtin(self) -> List[IntentCheck]:
        return [c for c in self.intent_checks if not c.custom]

    @property
    @loaded_status
    def intent_by_id(self) -> Dict[str, IntentCheck]:
        return {c.intent_id: c for c in self.intent_checks}

    @property
    @loaded_status
    def intents_by_name(self) -> Dict[str, List[IntentCheck]]:
        reports = defaultdict(list)
        [reports[c.name].append(c) for c in self.intent_checks]
        return dict(reports)

    @property
    @loaded_status
    def intent_by_name(self) -> Dict[str, IntentCheck]:
        msg = (
            "This method will be removed in v6.10/v7.0. "
            "Please use IPFClient().intent.intents_by_name() as Intent Names are not unique."
        )
        warn(
            msg,
            DeprecationWarning,
            stacklevel=2,
        )
        logger.warning(msg)
        return {c.name: c for c in self.intent_checks}

    @property
    @loaded_status
    def intent_ids_by_web_endpoint(self) -> Dict[str, List[str]]:
        reports = defaultdict(list)
        [reports[c.web_endpoint].append(c.intent_id) for c in self.intent_checks]
        return dict(reports)

    @property
    @loaded_status
    def intent_ids_by_api_endpoint(self) -> Dict[str, List[str]]:
        reports = defaultdict(list)
        [reports[c.api_endpoint].append(c.intent_id) for c in self.intent_checks]
        return dict(reports)

    @property
    @loaded_status
    def group_by_id(self) -> Dict[str, Group]:
        return {g.group_id: g for g in self.groups}

    @property
    @loaded_status
    def group_by_name(self) -> Dict[str, Group]:
        return {g.name: g for g in self.groups}

    @loaded_status
    def get_results(self, intent: IntentCheck, color: Union[str, int], snapshot_id: str = None) -> list:
        """Get the outcome of an Intent Check by a specific color

        Args:
            intent: an IntentCheck, please see the Intent Check Model
            color: color of intent check
            snapshot_id: Uses a different Snapshot ID then client

        Returns:
            list: List of Dictionary objects.
        """
        if isinstance(color, str):
            color = COLOR_DICT[color]
        return self._get_data(intent, snapshot_id or self.snapshot_id, color)

    @loaded_status
    def get_all_results(self, intent: IntentCheck, snapshot_id: str = None):
        """set the intent check attributes

        Args:
            intent: an IntentCheck, please see the Intent Check Model
            snapshot_id: Uses a different Snapshot ID then client

        Returns:
            list: List of Dictionary objects.
        """
        snapshot_id = snapshot_id or self.snapshot_id
        for color_str, color_int in COLOR_DICT.items():
            if color_int == -1 or getattr(intent.result.checks, color_str):
                setattr(intent.result_data, color_str, self._get_data(intent, snapshot_id, color_int))
        return intent

    def _get_data(self, intent: IntentCheck, snapshot_id: str, color: int):
        return self.client.fetch_all(
            intent.api_endpoint,
            snapshot_id=snapshot_id,
            reports=intent.web_endpoint,
            filters={intent.column: ["color", "eq", color]},
        )

    @loaded_status
    def compare_snapshot(self, snapshot_id: str, reverse: bool = False) -> list:
        """Compares all intents against another snapshot.
        Current is the snapshot loaded into the class
        Other is the snapshot specified in this method.  Use reverse=True to flip them.

        Args:
            snapshot_id: Snapshot ID to compare against this will be the "other" key
            reverse: Default False, setting to true will flip current and other.
        Returns:
            list: List of dictionaries
        """
        # TODO: Fix
        new_intents = {i.name: i for i in self.get_intent_checks(snapshot_id)}
        comparison = list()
        for name, intent in new_intents.items():
            old = self.intents_by_name[name][0].result
            compare = intent.result.compare(old) if reverse else old.compare(intent.result)
            for desc, value in compare.items():
                n = desc if desc != "count" else "total"
                comparison.append({"name": name, "id": intent.intent_id, "check": n, **value})
        return comparison



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class ManagedNetworks(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def networks(self):
        return models.Table(client=self.client, endpoint="tables/networks", sn=self.sn)

    @property
    def gateway_redundancy(self):
        return models.Table(client=self.client, endpoint="tables/networks/gateway-redundancy", sn=self.sn)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ipfabric import IPFClient
    from ipfabric.api import IPFabricAPI

import logging
from datetime import datetime
from time import sleep
from typing import Optional, List, Union, Set, Dict, Literal, Any
from httpx import HTTPError
from pathlib import Path
from ipfabric.models import Device
from ipfabric.tools import validate_ip_network_str, VALID_IP

from pydantic import BaseModel, Field, PrivateAttr
from .jobs import Jobs
from ipfabric.exceptions import deprecated_args_decorator

logger = logging.getLogger("ipfabric")

SNAPSHOT_COLUMNS = {
    "id",
    "creatorUsername",
    "status",
    "finishStatus",
    "loadedSize",
    "unloadedSize",
    "name",
    "note",
    "sites",
    "fromArchive",
    "loading",
    "locked",
    "deviceAddedCount",
    "deviceRemovedCount",
    "interfaceActiveCount",
    "interfaceCount",
    "interfaceEdgeCount",
    "totalDevCount",
    "isLastSnapshot",
    "tsChange",
    "tsEnd",
    "tsStart",
    "userCount",
}


def loaded_status(func):
    def _decorator(self, *args, **kwargs):
        if not self.loaded:
            logger.error(f"Snapshot {self.snapshot_id} is not loaded.")
            return False
        return func(self, *args, **kwargs)

    return _decorator


def snapshot_upload(ipf: IPFClient, filename: str):
    data = {"file": (Path(filename).name, open(filename, "rb"), "application/x-tar")}
    resp = ipf.post("snapshots/upload", files=data)
    if resp.status_code == 400:
        if resp.json()["code"] == "API_SNAPSHOT_CONFLICT":
            logger.warning(f"SNAPSHOT ID {resp.json()['data']['snapshot']} already uploaded")
            return
    resp.raise_for_status()
    return resp.json()


class Error(BaseModel):
    error_type: str = Field(None, alias="errorType")
    count: int


class Snapshot(BaseModel):
    client: Any
    snapshot_id: str = Field(None, alias="id")
    name: Optional[str] = None
    note: Optional[str] = None
    creator_username: Optional[str] = Field(None, alias="creatorUsername")
    total_dev_count: int = Field(None, alias="totalDevCount")
    licensed_dev_count: Optional[int] = Field(None, alias="licensedDevCount")
    user_count: int = Field(None, alias="userCount")
    interface_active_count: int = Field(None, alias="interfaceActiveCount")
    interface_count: int = Field(None, alias="interfaceCount")
    interface_edge_count: int = Field(None, alias="interfaceEdgeCount")
    device_added_count: int = Field(None, alias="deviceAddedCount")
    device_removed_count: int = Field(None, alias="deviceRemovedCount")
    status: str
    finish_status: str = Field(None, alias="finishStatus")
    loading: bool
    locked: bool
    from_archive: bool = Field(None, alias="fromArchive")
    start: datetime = Field(None, alias="tsStart")
    end: Optional[datetime] = Field(None, alias="tsEnd")
    change: Optional[datetime] = Field(None, alias="tsChange")
    version: Optional[str] = None
    initial_version: Optional[str] = Field(None, alias="initialVersion")
    sites: List[str]
    errors: Optional[List[Error]] = None
    loaded_size: int = Field(None, alias="loadedSize")
    unloaded_size: int = Field(None, alias="unloadedSize")
    disabled_graph_cache: Optional[bool] = None
    disabled_historical_data: Optional[bool] = None
    disabled_intent_verification: Optional[bool] = None
    __jobs: Jobs = PrivateAttr()

    @property
    def _jobs(self):
        if not getattr(self, "__jobs", None):
            self.__jobs = Jobs(client=self.client)
        return self.__jobs

    def discovery_errors(self, filters: dict = None, columns: List[str] = None, sort: dict = None) -> List[dict]:
        return self.client.fetch_all(
            "tables/reports/discovery-errors", snapshot_id=self.snapshot_id, filters=filters, columns=columns, sort=sort
        )

    def discovery_tasks(self, filters: dict = None, columns: List[str] = None, sort: dict = None) -> List[dict]:
        return self.client.fetch_all(
            "tables/reports/discovery-tasks", snapshot_id=self.snapshot_id, filters=filters, columns=columns, sort=sort
        )

    @property
    def errors_dict(self):
        return {_.error_type: _.count for _ in self.errors}

    @deprecated_args_decorator(version="6.9.0")
    @loaded_status
    def lock(self, ipf: IPFClient = None) -> bool:
        if not self.locked:
            res = self.client.post(f"snapshots/{self.snapshot_id}/lock")
            res.raise_for_status()
            self.locked = True
        else:
            logger.warning(f"Snapshot {self.snapshot_id} is already locked.")
        return True

    @deprecated_args_decorator(version="6.9.0")
    def unlock(self, ipf: IPFClient = None) -> bool:
        if self.locked and self.loaded:
            res = self.client.post(f"snapshots/{self.snapshot_id}/unlock")
            res.raise_for_status()
            self.locked = False
        else:
            logger.warning(f"Snapshot {self.snapshot_id} is already unlocked.")
        return True

    @property
    def loaded(self):
        return self.status == "done" and self.finish_status == "done"

    @deprecated_args_decorator(version="6.9.0", arg_type="IPFClient", no_args=False)
    def unload(self, ipf: IPFClient = None, wait_for_unload: bool = False, timeout: int = 60, retry: int = 5) -> bool:
        if not self.loaded:
            logger.warning(f"Snapshot {self.snapshot_id} is already unloaded.")
            return True
        ts = int(datetime.now().timestamp() * 1000)
        res = self.client.post("snapshots/unload", json=[dict(jobDetail=ts, id=self.snapshot_id)])
        res.raise_for_status()
        if wait_for_unload:
            if not self._jobs.check_snapshot_job(self.snapshot_id, ts, "unload", retry, timeout):
                logger.error("Snapshot Unload did not finish.")
                return False
        self._refresh_status()
        return True

    @deprecated_args_decorator(version="6.9.0", arg_type="IPFClient", no_args=False)
    def load(
        self,
        ipf: IPFClient = None,
        wait_for_load: bool = True,
        wait_for_assurance: bool = True,
        timeout: int = 60,
        retry: int = 5,
    ) -> bool:
        if self.loaded:
            logger.warning(f"Snapshot {self.snapshot_id} is already loaded.")
            return True
        ts = int(datetime.now().timestamp() * 1000)
        res = self.client.post("snapshots/load", json=[dict(jobDetail=ts, id=self.snapshot_id)])
        res.raise_for_status()
        if wait_for_load or wait_for_assurance:
            if not self._check_load_status(ts, wait_for_assurance, timeout, retry):
                return False
        self._refresh_status()
        return True

    def _refresh_status(self):
        results = self.client.fetch(
            "tables/management/snapshots",
            columns=["status", "finishStatus", "loading"],
            filters={"id": ["eq", self.snapshot_id]},
            snapshot=False,
        )[0]
        self.status, self.finish_status, self.loading = (
            results["status"],
            results["finishStatus"],
            results["loading"],
        )

    def _check_load_status(
        self,
        ts: int,
        wait_for_assurance: bool = True,
        timeout: int = 60,
        retry: int = 5,
    ):
        load_job = self._jobs.check_snapshot_job(
            self.snapshot_id, started=ts, action="load", timeout=timeout, retry=retry
        )
        if load_job and wait_for_assurance:
            return self._check_assurance_status(load_job["startedAt"], timeout, retry)
        elif not load_job:
            logger.error("Snapshot Load did not complete.")
            return False
        return True

    def _check_assurance_status(
        self,
        ts: int,
        timeout: int = 60,
        retry: int = 5,
    ):
        ae_settings = self.get_assurance_engine_settings(self.client)
        ae_status = False
        if ae_settings:
            ae_status = self._jobs.check_snapshot_assurance_jobs(
                self.snapshot_id, ae_settings, started=ts, timeout=timeout, retry=retry
            )
            if not ae_status:
                logger.error("Assurance Engine tasks did not complete")
        elif not ae_settings:
            logger.error("Could not get Assurance Engine tasks please check permissions.")
        if not ae_settings or not ae_status:
            self._change_snapshot()
            return False
        else:
            self.client.update()
            return True

    @deprecated_args_decorator(version="6.9.0")
    def attributes(self, ipf: IPFClient = None):
        """
        Load Snapshot
        :param ipf: IPFClient
        :return: True
        """
        return self.client.fetch_all("tables/snapshot-attributes", snapshot_id=self.snapshot_id)

    @deprecated_args_decorator(version="6.9.0", arg_type="IPFClient", no_args=False)
    def download(self, ipf: IPFClient = None, path: str = None, timeout: int = 60, retry: int = 5):
        if not path:
            path = Path(f"{self.snapshot_id}.tar")
        elif not isinstance(path, Path):
            path = Path(f"{path}")
        if not path.name.endswith(".tar"):
            path = Path(f"{path.name}.tar")

        # start download job
        ts = int(datetime.now().timestamp() * 1000)
        resp = self.client.get(f"/snapshots/{self.snapshot_id}/download")
        resp.raise_for_status()

        # waiting for download job to process
        job = self._jobs.check_snapshot_job(
            self.snapshot_id, started=ts, action="download", retry=retry, timeout=timeout
        )
        if job:
            filename = self.client.get(f"jobs/{job['id']}/download")
            with open(path, "wb") as fp:
                fp.write(filename.read())
            return path
        logger.error(f"Download job did not finish within {retry * timeout} seconds, could not get file.")
        return None

    @deprecated_args_decorator(version="6.9.0", arg_type=("IPFClient", "IPFabricAPI"), no_args=False)
    @loaded_status
    def get_snapshot_settings(self, ipf: Union[IPFClient, IPFabricAPI] = None) -> Union[dict, bool]:
        settings = None
        msg = "API_INSUFFICIENT_RIGHTS to `snapshots/:key/settings` " + self.client.user.error_msg

        if self.client.user.get_snapshots_settings is False:
            logger.debug(f"Could not get Snapshot {self.snapshot_id} Settings:" + msg)
            return settings
        res = self.client.get(f"/snapshots/{self.snapshot_id}/settings")
        try:
            res.raise_for_status()
            settings = res.json()
            if self.client.user.get_snapshots_settings is None:
                self.client.user.set_snapshots_settings(True)
        except HTTPError:
            logger.debug(f"Could not get Snapshot {self.snapshot_id} Settings:" + msg)
            logger.warning(msg)
            self.client.user.set_snapshots_settings(False)
        return settings

    @deprecated_args_decorator(version="6.9.0", arg_type=("IPFClient", "IPFabricAPI"), no_args=False)
    @loaded_status
    def get_assurance_engine_settings(self, ipf: Union[IPFClient, IPFabricAPI] = None) -> Union[dict, bool]:
        settings = self.get_snapshot_settings()
        if settings is None:
            logger.debug(f"Could not get Snapshot {self.snapshot_id} Settings to verify Assurance Engine tasks.")
            return False
        disabled = settings.get("disabledPostDiscoveryActions", list())
        self.disabled_graph_cache = True if "graphCache" in disabled else False
        self.disabled_historical_data = True if "historicalData" in disabled else False
        self.disabled_intent_verification = True if "intentVerification" in disabled else False
        return dict(
            disabled_graph_cache=self.disabled_graph_cache,
            disabled_historical_data=self.disabled_historical_data,
            disabled_intent_verification=self.disabled_intent_verification,
        )

    @deprecated_args_decorator(version="6.9.0", arg_type=("IPFClient", "IPFabricAPI"), no_args=False)
    @loaded_status
    def update_assurance_engine_settings(
        self,
        ipf: Union[IPFClient, IPFabricAPI] = None,
        disabled_graph_cache: bool = False,
        disabled_historical_data: bool = False,
        disabled_intent_verification: bool = False,
        wait_for_assurance: bool = True,
        timeout: int = 60,
        retry: int = 5,
    ) -> bool:
        settings = self.get_snapshot_settings(self.client)
        if settings is None:
            logger.error(
                f"Could not get Snapshot {self.snapshot_id} Settings and cannot update Assurance Engine tasks."
            )
            return False
        current = set(settings.get("disabledPostDiscoveryActions", list()))
        disabled, ae_settings = self._calculate_new_ae_settings(
            current, disabled_graph_cache, disabled_historical_data, disabled_intent_verification
        )
        if disabled == current:
            logger.info("No changes to Assurance Engine Settings required.")
            return True
        ts = int(datetime.now().timestamp() * 1000)
        res = self.client.patch(
            f"/snapshots/{self.snapshot_id}/settings", json=dict(disabledPostDiscoveryActions=list(disabled))
        )
        res.raise_for_status()
        if wait_for_assurance and current - disabled:
            ae_status = self._jobs.check_snapshot_assurance_jobs(
                self.snapshot_id, ae_settings, started=ts, timeout=timeout, retry=retry
            )
            if not ae_status:
                logger.error("Assurance Engine tasks did not complete")
                return False
        return True

    @staticmethod
    def _calculate_new_ae_settings(
        current: set,
        disabled_graph_cache: bool = False,
        disabled_historical_data: bool = False,
        disabled_intent_verification: bool = False,
    ):
        disabled = set()
        if disabled_graph_cache:
            disabled.add("graphCache")
        if disabled_historical_data:
            disabled.add("historicalData")
        if disabled_intent_verification:
            disabled.add("intentVerification")
        enabled = current - disabled

        ae_settings = dict(
            disabled_graph_cache=False if "graphCache" in enabled else True,
            disabled_historical_data=False if "historicalData" in enabled else True,
            disabled_intent_verification=False if "intentVerification" in enabled else True,
        )

        return disabled, ae_settings

    @staticmethod
    def _dev_to_sn(devices: Union[List[str], List[Device], Set[str], str, Device]) -> Set[str]:
        sns = set()
        if not isinstance(devices, (list, set)):
            devices = {devices}
        for device in devices:
            if isinstance(device, str):
                sns.add(device)
            elif isinstance(device, Device):
                sns.add(device.sn)
        return sns

    @loaded_status
    def verify_snapshot_devices(
        self, devices: Union[List[str], List[Device], Set[str], str, Device]
    ) -> Dict[str, Set[str]]:
        """Checks to ensure that the Vendor is enabled for the devices.

        Args:
            devices: IP Fabric Device Serial Number, Device object, or list of either

        Returns: bool
        """
        sn = self._dev_to_sn(devices)
        payload = {
            "columns": ["id", "isApiTask", "sn", "settingsStates", "vendor"],
            "filters": {"isSelected": ["eq", True]},
            "bindVariables": {"selected": list(sn), "isDelete": False},
            "snapshot": self.snapshot_id,
        }
        snap_devs = self.client._ipf_pager("tables/snapshot-devices", payload)
        disabled = {dev["sn"] for dev in snap_devs if "noApiSettings" in dev["settingsStates"] and dev["isApiTask"]}
        valid = {dev["sn"] for dev in snap_devs if "ok" in dev["settingsStates"]}
        invalid = sn - valid - disabled
        if disabled:
            logger.warning(f"Vendor API(s) disabled for devices: {list(disabled)}.")
        if invalid:
            logger.warning(f"Invalid snapshot devices: {list(invalid)}.")
        return {"disabled": disabled, "invalid": invalid, "valid": valid}

    @loaded_status
    def delete_devices(
        self,
        devices: Union[List[str], List[Device], Set[str], str, Device],
        wait_for_discovery: bool = True,
        wait_for_assurance: bool = True,
        timeout: int = 60,
        retry: int = 5,
        skip_invalid_devices: bool = True,
    ) -> bool:
        """Rediscover device(s) based on a serial number or Device object.

        Args:
            devices: IP Fabric Device Serial Number, Device object, or list of either
            wait_for_discovery: Default True to wait before returning
            wait_for_assurance:  Default True to wait before returning
            timeout: How long to wait, larger snapshots will take longer times
            retry: Number of retries to wait ( (timeout X retry = total length) X 2) Will check for Discovery
                   and then use same times for Assurance.
            skip_invalid_devices: If a Vendor API device is included and the vendor is disabled it will be skipped
                                  instead of not refreshing devices.

        Returns: bool
        """
        return self._rediscover_delete_devices(
            "delete", devices, wait_for_discovery, wait_for_assurance, timeout, retry, skip_invalid_devices
        )

    @loaded_status
    def rediscover_devices(
        self,
        devices: Union[List[str], List[Device], Set[str], str, Device],
        wait_for_discovery: bool = True,
        wait_for_assurance: bool = True,
        timeout: int = 60,
        retry: int = 5,
        skip_invalid_devices: bool = True,
    ) -> bool:
        """Rediscover device(s) based on a serial number or Device object.

        Args:
            devices: IP Fabric Device Serial Number, Device object, or list of either
            wait_for_discovery: Default True to wait before returning
            wait_for_assurance:  Default True to wait before returning
            timeout: How long to wait, larger snapshots will take longer times
            retry: Number of retries to wait ( (timeout X retry = total length) X 2) Will check for Discovery
                   and then use same times for Assurance.
            skip_invalid_devices: If a Vendor API device is included and the vendor is disabled it will be skipped
                                  instead of not refreshing devices.

        Returns: bool
        """
        return self._rediscover_delete_devices(
            "refresh", devices, wait_for_discovery, wait_for_assurance, timeout, retry, skip_invalid_devices
        )

    @loaded_status
    def _rediscover_delete_devices(
        self,
        action: Literal["refresh", "delete"],
        devices: Union[List[str], List[Device], Set[str], str, Device],
        wait_for_discovery: bool = True,
        wait_for_assurance: bool = True,
        timeout: int = 60,
        retry: int = 5,
        skip_invalid_devices: bool = True,
    ) -> bool:
        """Rediscover device(s) based on a serial number or Device object.

        Args:
            devices: IP Fabric Device Serial Number, Device object, or list of either
            wait_for_discovery: Default True to wait before returning
            wait_for_assurance:  Default True to wait before returning
            timeout: How long to wait, larger snapshots will take longer times
            retry: Number of retries to wait ( (timeout X retry = total length) X 2) Will check for Discovery
                   and then use same times for Assurance.
            skip_invalid_devices: If a Vendor API device is included and the vendor is disabled it will be skipped
                                  instead of not refreshing devices.

        Returns: bool
        """
        sn = self._dev_to_sn(devices)
        devices = self.verify_snapshot_devices(sn)
        if not skip_invalid_devices and (devices["disabled"] or devices["invalid"]):
            return False
        sn = list(devices["valid"])
        if not sn:
            return False

        ts = int(datetime.now().timestamp() * 1000)
        if action == "delete":
            resp = self.client.request("DELETE", f"snapshots/{self.snapshot_id}/devices", json=sn)
        else:
            resp = self.client.post(
                f"snapshots/{self.snapshot_id}/devices", json=dict(snList=sn, vendorSettingsMap=dict())
            )
        resp.raise_for_status()
        if not wait_for_discovery:
            self._change_snapshot()
            return resp.json()["success"]

        return self._check_modification_status(wait_for_assurance, ts, timeout, retry, action)

    def _vendor_apis(self):
        vendors = list()
        for vendor in self.client.get(f"snapshots/{self.snapshot_id}/available-vendor-settings").json():
            vendor.pop("details", None)
            if vendor["type"] not in ["juniper-mist", "ruckus-vsz"]:
                vendor.pop("apiVersion", None)
            if vendor["type"] == "aws-ec2":
                vendor.pop("baseUrl", None)
            vendors.append(vendor)
        return vendors

    @loaded_status
    def add_devices(
        self,
        ip: Union[List[str], List[VALID_IP], str, VALID_IP] = None,
        refresh_vendor_api: bool = True,
        retry_timed_out: bool = True,
        wait_for_discovery: bool = True,
        wait_for_assurance: bool = True,
        timeout: int = 60,
        retry: int = 5,
    ) -> bool:
        """Add device(s) based on a IP address or subnet.

        Args:
            ip: Single IP or list of IPs
            refresh_vendor_api: Default True to refresh Vendor API devices that are enabled in snapshot settings.
            retry_timed_out: IP Default True to retry devices that timed out.
            wait_for_discovery: Default True to wait before returning
            wait_for_assurance:  Default True to wait before returning
            timeout: How long to wait, larger snapshots will take longer times
            retry: Number of retries to wait ( (timeout X retry = total length) X 2) Will check for Discovery
                   and then use same times for Assurance.

        Returns: bool
        """
        if not ip and not refresh_vendor_api and not retry_timed_out:
            raise SyntaxError("No snapshot modification selected.")
        elif not ip and not refresh_vendor_api and not self.errors_dict.get("ABCommandTimeout", 0):
            logger.warning(f"No Command Timeout Errors found in {self.snapshot_id}, not refreshing snapshot.")
            return True
        vendors = self._vendor_apis() if refresh_vendor_api else list()
        ips = list()
        ip = list() if ip is None else ip
        for i in [ip] if isinstance(ip, str) else ip:
            i = validate_ip_network_str(i)
            if int(i.split("/")[1]) < 23:
                raise ValueError(f"IP Network {i} is larger than /23.")
            ips.append(i)

        payload = {"ipList": ips, "retryTimedOut": retry_timed_out, "vendorApi": vendors}

        ts = int(datetime.now().timestamp() * 1000)
        resp = self.client.post(f"snapshots/{self.snapshot_id}/devices", json=payload)
        resp.raise_for_status()
        if not wait_for_discovery:
            self._change_snapshot()
            return resp.json()["success"]
        return self._check_modification_status(wait_for_assurance, ts, timeout, retry, "add")

    def _change_snapshot(self):
        logger.warning(f"Snapshot {self.snapshot_id} is discovering switching to $last.")
        sleep(2)
        self.client.update()

    def _check_modification_status(self, wait_for_assurance, ts, timeout, retry, action):
        job = self._jobs.check_snapshot_job(self.snapshot_id, started=ts, action=action, timeout=timeout, retry=retry)
        if job and wait_for_assurance:
            return self._check_assurance_status(job["startedAt"], timeout, retry)
        elif not job:
            logger.error(f"Snapshot Discovery {action.capitalize()} did not complete.")
        self.client.update()
        return True if job else False

    @loaded_status
    def add_ip_devices(
        self,
        ip: Union[List[str], List[VALID_IP], str, VALID_IP],
        wait_for_discovery: bool = True,
        wait_for_assurance: bool = True,
        timeout: int = 60,
        retry: int = 5,
    ) -> bool:
        """Refreshes Vendor API devices.

        Args:
            ip: Single IP or list of IPs
            wait_for_discovery: Default True to wait before returning
            wait_for_assurance:  Default True to wait before returning
            timeout: How long to wait, larger snapshots will take longer times
            retry: Number of retries to wait ( (timeout X retry = total length) X 2) Will check for Discovery
                   and then use same times for Assurance.

        Returns: bool
        """
        return self.add_devices(ip, False, False, wait_for_discovery, wait_for_assurance, timeout, retry)

    @loaded_status
    def refresh_vendor_api(
        self,
        wait_for_discovery: bool = True,
        wait_for_assurance: bool = True,
        timeout: int = 60,
        retry: int = 5,
    ) -> bool:
        """Refreshes Vendor API devices.

        Args:
            wait_for_discovery: Default True to wait before returning
            wait_for_assurance:  Default True to wait before returning
            timeout: How long to wait, larger snapshots will take longer times
            retry: Number of retries to wait ( (timeout X retry = total length) X 2) Will check for Discovery
                   and then use same times for Assurance.

        Returns: bool
        """
        return self.add_devices(None, True, False, wait_for_discovery, wait_for_assurance, timeout, retry)

    @loaded_status
    def retry_timed_out_devices(
        self,
        wait_for_discovery: bool = True,
        wait_for_assurance: bool = True,
        timeout: int = 60,
        retry: int = 5,
    ) -> bool:
        """Retries devices that timed out in Snapshot.

        Args:
            wait_for_discovery: Default True to wait before returning
            wait_for_assurance:  Default True to wait before returning
            timeout: How long to wait, larger snapshots will take longer times
            retry: Number of retries to wait ( (timeout X retry = total length) X 2) Will check for Discovery
                   and then use same times for Assurance.

        Returns: bool
        """
        return self.add_devices(None, False, True, wait_for_discovery, wait_for_assurance, timeout, retry)



import logging
from ipaddress import IPv4Interface, AddressValueError
from typing import Any

from pydantic.dataclasses import dataclass

logger = logging.getLogger("ipfabric")


@dataclass
class Seeds:
    client: Any

    @property
    def seeds(self):
        res = self.client.get("settings/seed")
        res.raise_for_status()
        return res.json()

    @staticmethod
    def _check_seeds(seeds):
        valid = True
        for seed in seeds:
            try:
                IPv4Interface(seed)
            except AddressValueError:
                valid = False
                logger.error(f"Seed {seed} is not a valid IP address or network.")
        return valid

    def set_seeds(self, seeds):
        """
        Sets the seeds with supplied list, will override current configuration.
        :param seeds: list: List of IP addresses or networks
        :return: list: Updated list of configured seeds
        """
        if not isinstance(seeds, list):
            raise SyntaxError("Seeds must be a list of IP Addresses or networks")
        if self._check_seeds(seeds):
            res = self.client.put("settings/seed", json=seeds)
            res.raise_for_status()
            return res.json()

    def add_seeds(self, seeds):
        """
        Adds a single seed or list of seeds to the current configuration
        :param seeds: list: List of IP addresses or networks
        :return: list: Updated list of configured seeds
        """
        if isinstance(seeds, str):
            seeds = [seeds]
        updated = list(set.union(set(self.seeds), set(seeds)))
        return self.set_seeds(updated)

    def delete_seeds(self, seeds):
        """
        Deletes a seed or a list of seeds from what is configured.
        :param seeds: list: List of IP addresses or networks
        :return: list: Updated list of configured seeds
        """
        if isinstance(seeds, str):
            seeds = [seeds]
        updated = list(set(self.seeds) - set(seeds))
        return self.set_seeds(updated)



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Neighbors(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def neighbors_all(self):
        return models.Table(client=self.client, endpoint="tables/neighbors/all", sn=self.sn)

    @property
    def neighbors_unmanaged(self):
        return models.Table(client=self.client, endpoint="tables/neighbors/unmanaged", sn=self.sn)

    @property
    def neighbors_unidirectional(self):
        return models.Table(client=self.client, endpoint="tables/neighbors/unidirectional", sn=self.sn)

    @property
    def neighbors_endpoints(self):
        return models.Table(client=self.client, endpoint="tables/neighbors/endpoints", sn=self.sn)



from time import sleep
from typing import List, Optional

from httpx import Client, ReadTimeout, HTTPStatusError
from httpx._types import ProxiesTypes
from pydantic import BaseModel


class CVE(BaseModel):
    cve_id: str
    description: str
    url: str

    def __repr__(self):
        return self.cve_id

    def __hash__(self):
        return hash(self.cve_id)


class CVEs(BaseModel):
    total_results: int
    cves: List[CVE]
    error: Optional[str] = None


class NIST(Client):
    def __init__(self, nvd_api_key: str, timeout, proxies: Optional[ProxiesTypes] = None):
        """
        NIST updated to API v2.0.  You must request and pass an API Key which can be obtained at
        https://nvd.nist.gov/developers/request-an-api-key

        Args:
            nvd_api_key: str: https://nvd.nist.gov/developers/request-an-api-key
            timeout: int:
        """
        super().__init__(
            base_url="https://services.nvd.nist.gov/rest/json/cves/2.0",
            timeout=timeout,
            headers={"apiKey": nvd_api_key},
            proxy=proxies,
        )

    @property
    def params(self):
        return {"virtualMatchString": "cpe:2.3:*:", "startIndex": 0}

    @staticmethod
    def _check_cisco(family, version, vendor, params):
        if family == "wlc-air":
            family = "wireless_lan_controller_software"
        elif family == "ftd":
            family = "firepower"
            version = (version.replace("(Bu", ".Bu")).replace(")", ".").split(" .", 1)[0]
        elif family != "nx-os":
            family = family.replace("-", "_")
        version = (version.replace("(", ".")).replace(")", ".")
        params["virtualMatchString"] += vendor + ":" + family + ":" + version
        return params

    def check_cve(self, vendor: str, family: str, version: str) -> CVEs:  # noqa: C901
        """returns CVE data about a specific product line

        Args:
        vendor: Vendor for the device to be checked
        family: Family of the device to be checked
        version: Software version of the device to be checked
        Returns:
            a list of CVEs
        """
        params = self.params
        if vendor in ["azure", "aws"] or family == "meraki":
            return CVEs(total_results=0, cves=[], error="Unsupported")
        elif vendor == "juniper":
            version = version[: version.rfind("R") + 2].replace("R", ":r")
            params["virtualMatchString"] += vendor + ":" + family + ":" + version
        elif vendor == "paloalto":
            params["virtualMatchString"] += "palo_alto" + ":" + family + ":" + version
        elif vendor == "extreme":
            family = "extremexos" if "xos" in family else family
            params["virtualMatchString"] += "extremenetworks" + ":" + family + ":" + version
        elif family and "aruba" in family:
            params["virtualMatchString"] += "arubanetworks:arubaos" + ":" + version
        elif vendor == "f5" and family == "big-ip":
            params["virtualMatchString"] += vendor + ":" + "big-ip_access_policy_manager" + ":" + version
        elif vendor == "cisco":
            params = self._check_cisco(family, version, vendor, params)
        elif vendor == "fortinet" and family == "fortigate":
            params["virtualMatchString"] += "fortinet:fortios:" + version.replace(",", ".")
        elif vendor == "checkpoint" and family == "gaia":
            params["virtualMatchString"] += vendor + ":" + "gaia_os" + version.replace("R", ":r")
        elif vendor == "arista":
            params["virtualMatchString"] += vendor + ":" + family + ":" + version.lower()
        else:
            return CVEs(total_results=0, cves=[], error="Unsupported")
        return self._query_cve(params)

    def _query_cve(self, params):
        try:
            sleep(0.3)  # NIST Rate Limit: 50 requests/30 seconds
            res = self.get("", params=params)
            res.raise_for_status()
            data = res.json()

            cves = CVEs(
                total_results=data["totalResults"],
                cves=[
                    CVE(
                        cve_id=i["cve"]["id"],
                        description=i["cve"]["descriptions"][0]["value"],
                        url=i["cve"]["references"][0]["url"],
                    )
                    for i in data["vulnerabilities"]
                ],
            )
            return cves
        except ReadTimeout:
            return CVEs(total_results=0, cves=[], error="Timeout")
        except HTTPStatusError:
            return CVEs(total_results=0, cves=[], error="HTTP Error")



from .input_models.graph_parameters import ICMP

ECHO_REPLY = ICMP(type=0, code=0)
NET_UNREACHABLE = ICMP(type=3, code=0)
HOST_UNREACHABLE = ICMP(type=3, code=1)
PROTOCOL_UNREACHABLE = ICMP(type=3, code=2)
PORT_UNREACHABLE = ICMP(type=3, code=3)
FRAGMENTATION_NEEDED = ICMP(type=3, code=4)
SOURCE_ROUTE_FAILED = ICMP(type=3, code=5)
DESTINATION_NETWORK_UNKNOWN = ICMP(type=3, code=6)
DESTINATION_HOST_UNKNOWN = ICMP(type=3, code=7)
SOURCE_HOST_ISOLATED = ICMP(type=3, code=8)
DESTINATION_NETWORK_ADMIN_PROHIBITED = ICMP(type=3, code=9)
DESTINATION_HOST_ADMIN_PROHIBITED = ICMP(type=3, code=10)
DESTINATION_NETWORK_UNREACHABLE_FOR_TOS = ICMP(type=3, code=11)
DESTINATION_HOST_UNREACHABLE_FOR_TOS = ICMP(type=3, code=12)
COMMUNICATION_ADMINISTRATIVELY_PROHIBITED = ICMP(type=3, code=13)
HOST_PRECEDENCE_VIOLATION = ICMP(type=3, code=14)
PRECEDENCE_CUTOFF_IN_EFFECT = ICMP(type=3, code=15)
SOURCE_QUENCH = ICMP(type=4, code=0)
REDIRECT_DATAGRAM_FOR_THE_NETWORK = ICMP(type=5, code=0)
REDIRECT_DATAGRAM_FOR_THE_HOST = ICMP(type=5, code=1)
REDIRECT_DATAGRAM_FOR_THE_TOS_AND_NETWORK = ICMP(type=5, code=2)
REDIRECT_DATAGRAM_FOR_THE_TOS_AND_HOST = ICMP(type=5, code=3)
ALTERNATE_ADDRESS_FOR_HOST = ICMP(type=6, code=0)
ECHO_REQUEST = ICMP(type=8, code=0)
NORMAL_ROUTER_ADVERTISEMENT = ICMP(type=9, code=0)
DOES_NOT_ROUTE_COMMON_TRAFFIC = ICMP(type=9, code=16)
ROUTER_SOLICITATION = ICMP(type=10, code=0)
TIME_TO_LIVE_EXCEEDED_IN_TRANSIT = ICMP(type=11, code=0)
FRAGMENT_REASSEMBLY_TIME_EXCEEDED = ICMP(type=11, code=1)
POINTER_INDICATES_THE_ERROR = ICMP(type=12, code=0)
MISSING_A_REQUIRED_OPTION = ICMP(type=12, code=1)
BAD_LENGTH = ICMP(type=12, code=2)
TIMESTAMP_REQUEST = ICMP(type=13, code=0)
TIMESTAMP_REPLY = ICMP(type=14, code=0)
INFORMATION_REQUEST = ICMP(type=15, code=0)
INFORMATION_REPLY = ICMP(type=16, code=0)
MASK_REQUEST = ICMP(type=17, code=0)
MASK_REPLY = ICMP(type=18, code=0)
BAD_SPI = ICMP(type=40, code=0)
AUTHENTICATION_FAILED = ICMP(type=40, code=1)
DECOMPRESSION_FAILED = ICMP(type=40, code=2)
DECRYPTION_FAILED = ICMP(type=40, code=3)
NEED_AUTHENTICATION = ICMP(type=40, code=4)
NEED_AUTHORIZATION = ICMP(type=40, code=5)

__all__ = [
    "ICMP",
    "ECHO_REPLY",
    "NET_UNREACHABLE",
    "HOST_UNREACHABLE",
    "PROTOCOL_UNREACHABLE",
    "PORT_UNREACHABLE",
    "FRAGMENTATION_NEEDED",
    "SOURCE_ROUTE_FAILED",
    "DESTINATION_NETWORK_UNKNOWN",
    "DESTINATION_HOST_UNKNOWN",
    "SOURCE_HOST_ISOLATED",
    "DESTINATION_NETWORK_ADMIN_PROHIBITED",
    "DESTINATION_HOST_ADMIN_PROHIBITED",
    "DESTINATION_NETWORK_UNREACHABLE_FOR_TOS",
    "DESTINATION_HOST_UNREACHABLE_FOR_TOS",
    "COMMUNICATION_ADMINISTRATIVELY_PROHIBITED",
    "HOST_PRECEDENCE_VIOLATION",
    "PRECEDENCE_CUTOFF_IN_EFFECT",
    "SOURCE_QUENCH",
    "REDIRECT_DATAGRAM_FOR_THE_NETWORK",
    "REDIRECT_DATAGRAM_FOR_THE_HOST",
    "REDIRECT_DATAGRAM_FOR_THE_TOS_AND_NETWORK",
    "REDIRECT_DATAGRAM_FOR_THE_TOS_AND_HOST",
    "ALTERNATE_ADDRESS_FOR_HOST",
    "ECHO_REQUEST",
    "NORMAL_ROUTER_ADVERTISEMENT",
    "DOES_NOT_ROUTE_COMMON_TRAFFIC",
    "ROUTER_SOLICITATION",
    "TIME_TO_LIVE_EXCEEDED_IN_TRANSIT",
    "FRAGMENT_REASSEMBLY_TIME_EXCEEDED",
    "POINTER_INDICATES_THE_ERROR",
    "MISSING_A_REQUIRED_OPTION",
    "BAD_LENGTH",
    "TIMESTAMP_REQUEST",
    "TIMESTAMP_REPLY",
    "INFORMATION_REQUEST",
    "INFORMATION_REPLY",
    "MASK_REQUEST",
    "MASK_REPLY",
    "BAD_SPI",
    "AUTHENTICATION_FAILED",
    "DECOMPRESSION_FAILED",
    "DECRYPTION_FAILED",
    "NEED_AUTHENTICATION",
    "NEED_AUTHORIZATION",
]



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class LoadBalancing(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def virtual_servers(self):
        return models.Table(client=self.client, endpoint="tables/load-balancing/virtual-servers", sn=self.sn)

    @property
    def virtual_servers_pools(self):
        return models.Table(client=self.client, endpoint="tables/load-balancing/virtual-servers/pools", sn=self.sn)

    @property
    def virtual_servers_pool_members(self):
        return models.Table(
            client=self.client, endpoint="tables/load-balancing/virtual-servers/pool-members", sn=self.sn
        )

    @property
    def virtual_servers_f5_partitions(self):
        return models.Table(client=self.client, endpoint="tables/load-balancing/f5-partitions", sn=self.sn)



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class PortChannels(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def inbound_balancing_table(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/port-channel/balance/inbound", sn=self.sn)

    @property
    def outbound_balancing_table(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/port-channel/balance/outbound", sn=self.sn)

    @property
    def member_status_table(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/port-channel/member-status", sn=self.sn)

    @property
    def mlag_switches(self):
        return models.Table(client=self.client, endpoint="tables/platforms/mlag/switches", sn=self.sn)

    @property
    def mlag_peers(self):
        return models.Table(client=self.client, endpoint="tables/platforms/mlag/peers", sn=self.sn)

    @property
    def mlag_pairs(self):
        return models.Table(client=self.client, endpoint="tables/platforms/mlag/pairs", sn=self.sn)

    @property
    def mlag_cisco_vpc(self):
        return models.Table(client=self.client, endpoint="tables/platforms/mlag/vpc", sn=self.sn)



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Multicast(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def pim_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/multicast/pim/neighbors", sn=self.sn)

    @property
    def pim_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/multicast/pim/interfaces", sn=self.sn)

    @property
    def mroute_overview(self):
        return models.Table(client=self.client, endpoint="tables/multicast/routes/overview", sn=self.sn)

    @property
    def mroute_table(self):
        return models.Table(client=self.client, endpoint="tables/multicast/routes/table", sn=self.sn)

    @property
    def mroute_oil_detail(self):
        return models.Table(client=self.client, endpoint="tables/multicast/routes/outgoing-interfaces", sn=self.sn)

    @property
    def mroute_counters(self):
        return models.Table(client=self.client, endpoint="tables/multicast/routes/counters", sn=self.sn)

    @property
    def mroute_first_hop_router(self):
        return models.Table(client=self.client, endpoint="tables/multicast/routes/first-hop-router", sn=self.sn)

    @property
    def mroute_sources(self):
        return models.Table(client=self.client, endpoint="tables/multicast/routes/sources", sn=self.sn)

    @property
    def igmp_groups(self):
        return models.Table(client=self.client, endpoint="tables/multicast/igmp/groups", sn=self.sn)

    @property
    def igmp_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/multicast/igmp/interfaces", sn=self.sn)

    @property
    def igmp_snooping_global_config(self):
        return models.Table(client=self.client, endpoint="tables/multicast/igmp/snooping/global", sn=self.sn)

    @property
    def igmp_snooping_groups(self):
        return models.Table(client=self.client, endpoint="tables/multicast/igmp/snooping/groups", sn=self.sn)

    @property
    def igmp_snooping_vlans(self):
        return models.Table(client=self.client, endpoint="tables/multicast/igmp/snooping/vlans", sn=self.sn)

    @property
    def mac_table(self):
        return models.Table(client=self.client, endpoint="tables/multicast/mac", sn=self.sn)

    @property
    def rp_overview(self):
        return models.Table(client=self.client, endpoint="tables/multicast/pim/rp/overview", sn=self.sn)

    @property
    def rp_bsr(self):
        return models.Table(client=self.client, endpoint="tables/multicast/pim/rp/bsr", sn=self.sn)

    @property
    def rp_mappings(self):
        return models.Table(client=self.client, endpoint="tables/multicast/pim/rp/mappings", sn=self.sn)

    @property
    def rp_mappings_groups(self):
        return models.Table(client=self.client, endpoint="tables/multicast/pim/rp/mappings-groups", sn=self.sn)
